from dataclasses import dataclass
from graphlib import TopologicalSorter

import graphviz
import networkx as nx

from vulkan.spec.dependency import INPUT_NODE, Dependency
from vulkan.spec.nodes.base import Node, NodeType

GraphNodes = list[Node]
GraphEdges = dict[str, dict[str, Dependency]]
"""Mapping of node names to their dependencies"""
Dependencies = dict[str, Dependency]
"""Map of a node's input variables to dependencies on other nodes' outputs."""


@dataclass
class GraphDefinition:
    """Specifies a graph via the nodes and the dependencies between them.

    Internal convenience used to share common functionality between different
    types of graphs.
    Should not be used directly.
    """

    nodes: GraphNodes

    def __post_init__(self):
        if len(self.nodes) == 0:
            raise ValueError("Policy must have at least one node")
        if not all(isinstance(n, Node) for n in self.nodes):
            msg = f"All elements must be of type Node, got {self.nodes}"
            raise ValueError(msg)

        self._validate_nodes()
        self._validate_node_dependencies()
        self._graph = _digraph_from_nodes(self.nodes)
        valid, errors = self._validate_graph()
        self._valid = valid
        self._errors = errors

        # self._node_definitions = {n.id: n.node_definition() for n in self.nodes}
        self._dependency_definitions = {n.id: n.dependencies for n in self.nodes}

    @property
    def edges(self) -> GraphEdges:
        return self._dependency_definitions

    # @property
    # def node_definitions(self) -> dict[str, NodeDefinition]:
    #     return self._node_definitions

    @property
    def valid(self) -> bool:
        return self._valid

    @property
    def errors(self) -> list[str]:
        return self._errors

    @property
    def leaves(self) -> dict[str, Node]:
        return {
            n: self._graph.nodes[n]["node"] for n, d in self._graph.out_degree if d == 0
        }

    def _validate_node_dependencies(self):
        node_dependencies = {node.id: node.dependencies for node in self.nodes}
        for node_name, dependencies in node_dependencies.items():
            if dependencies is None:
                continue

            for dep in dependencies.values():
                if dep.node == INPUT_NODE:
                    # Input nodes are added to the graph after validation.
                    continue

                if dep.id not in node_dependencies.keys():
                    msg = (
                        f"Node {node_name} has a dependency {dep.id} "
                        "that is not in the graph"
                    )
                    raise ValueError(msg)

    def show(self):
        return nx_to_graphviz(self._graph)

    def _validate_graph(self) -> tuple[bool, list[str]]:
        errors = []
        if not nx.is_directed_acyclic_graph(self._graph):
            errors.append("Policy graph contains cycles")

        terminate_nodes = [
            node for node in self.nodes if node.type == NodeType.TERMINATE
        ]
        if len(terminate_nodes) == 0:
            errors.append("No terminate node found in policy.")

        leaves = {
            name: leaf.type == NodeType.TERMINATE for name, leaf in self.leaves.items()
        }
        if not all(leaves.values()):
            msg = f"Policy leaves must be {NodeType.TERMINATE} nodes: {leaves}"
            errors.append(msg)

        return (len(errors) == 0), errors

    def _validate_nodes(self):
        for node in self.nodes:
            if not isinstance(node, Node):
                raise ValueError("All elements must be of type `Node`")

        ids = [node.id for node in self.nodes]
        duplicates = set([node_id for node_id in ids if ids.count(node_id) > 1])

        if duplicates:
            raise ValueError(f"Duplicate node names found: {duplicates}")


def nx_to_graphviz(G, prog="dot"):
    """
    Convert a NetworkX DiGraph to a Graphviz visualization
    Parameters:
    - G: NetworkX DiGraph
    - prog: Graphviz layout program ('dot', 'neato', 'fdp', 'sfdp', 'twopi', 'circo')
    Returns:
    - Graphviz object
    """
    # Create a new Graphviz Digraph
    dot = graphviz.Digraph(engine=prog)

    # Add nodes
    for node in G.nodes():
        # Get node attributes if any
        attrs = G.nodes[node]
        # Convert all attribute values to strings
        attrs = {k: str(v) for k, v in attrs.items()}
        dot.node(str(node), **attrs)

    # Add edges
    for u, v in G.edges():
        # Get edge attributes if any
        attrs = G.edges[u, v]
        # Convert all attribute values to strings
        attrs = {k: str(v) for k, v in attrs.items()}
        dot.edge(str(u), str(v), **attrs)

    return dot


def _digraph_from_nodes(nodes: list[Node]) -> nx.DiGraph:
    g = nx.DiGraph()
    g.add_nodes_from([(n.id, {"node": n}) for n in nodes])

    edges: list[tuple[str, str, dict]] = []
    for node in nodes:
        for dep_name, dep_node in node.dependencies.items():
            properties = {
                "dependency_name": dep_name,
                "source_node_name": dep_node.node,
                "source_node_output": dep_node.output,
                "source_node_key": dep_node.key,
                "source_node_hierarchy": dep_node.hierarchy,
            }
            e = (dep_node.id, node.id, properties)
            edges.append(e)

    g.add_edges_from(edges)
    return g


def extract_node_definitions(nodes: list[Node]) -> dict:
    return {node.id: node.to_dict() for node in nodes}


def sort_nodes(nodes: GraphNodes, edges: GraphEdges) -> GraphNodes:
    """Sort nodes topologically based on their dependencies."""
    node_map = {node.id: node for node in nodes}
    graph = {
        _id: [dep.id for dep in dependencies.values()]
        for _id, dependencies in edges.items()
    }
    sorter = TopologicalSorter(graph)

    # Input nodes may not be included in the nodes list, but will be
    # referenced as dependencies in the edges list.
    return [node_map[_id] for _id in sorter.static_order() if _id in node_map]



import json
from argparse import ArgumentParser

import apache_beam as beam
from apache_beam.options.pipeline_options import PipelineOptions
from apache_beam.pipeline import Pipeline

from vulkan.core.backtest.metrics import MetricsMetadata, Target, TargetKind
from vulkan.runners.beam.io import ReadParquet
from vulkan.runners.beam.metrics.binary import BinaryDistributionTransform


def build_metrics_pipeline(
    input_data_path: str,
    results_data_path: str,
    output_path: str,
    metrics_metadata: MetricsMetadata,
    pipeline_options: PipelineOptions,
) -> Pipeline:
    group_by_cols = [
        "backfill_id",
        metrics_metadata.outcome,
    ]
    if metrics_metadata.time is not None:
        group_by_cols.append(metrics_metadata.time)
    if metrics_metadata.groups is not None:
        group_by_cols.extend(metrics_metadata.groups)

    metrics_transform = _make_metrics_transform(
        target=metrics_metadata.target,
        group_by_cols=group_by_cols,
    )

    p = beam.Pipeline(options=pipeline_options)
    input_data = p | "Read Input Data" >> ReadParquet(input_data_path)
    results_data = (
        p
        | "Read Results Data" >> beam.io.ReadFromParquet(results_data_path)
        | beam.Map(lambda x: (x["key"], x))
    )
    data = {
        "results_data": results_data,
        "input_data": input_data,
    } | "Join Data" >> beam.CoGroupByKey()

    (
        data
        | "Select Relevant Columns"
        >> ExtractRelevantColumns(
            backfill_id_column="backfill_id",
            outcome_column=metrics_metadata.outcome,
            target_column=metrics_metadata.target.name,
            time_column=metrics_metadata.time,
            group_by_columns=metrics_metadata.groups,
        )
        | "Calculate Metrics" >> metrics_transform
        | "Enforce Schema" >> beam.Select(*metrics_transform.output_columns())
        | "Write Metrics"
        >> beam.io.WriteToJson(
            output_path,
            num_shards=1,
            file_naming=beam.io.fileio.single_file_naming(
                prefix="metrics", suffix=".json"
            ),
        )
    )

    return p


class ExtractRelevantColumns(beam.PTransform):
    def __init__(
        self,
        backfill_id_column: str,
        outcome_column: str,
        target_column: str,
        time_column: str | None,
        group_by_columns: list[str] | None,
    ):
        self.backfill_id_column = backfill_id_column
        self.target_column = target_column
        self.outcome_column = outcome_column
        self.time_column = time_column
        self.group_by_columns = group_by_columns

        _input_data_columns = [target_column]
        if time_column is not None:
            _input_data_columns.append(time_column)
        if group_by_columns is not None:
            _input_data_columns.extend(group_by_columns)
        self._input_data_columns = _input_data_columns

    def expand(self, pcoll):
        return pcoll | "Create Working Frame" >> beam.FlatMap(
            _extract_columns_by_key,
            input_data_columns=self._input_data_columns,
            outcome_column=self.outcome_column,
            backfill_id_column=self.backfill_id_column,
        )


def _extract_columns_by_key(
    element: tuple,
    input_data_columns: list[str],
    outcome_column: str,
    backfill_id_column: str,
):
    key = element[0]
    data = element[1]

    input_data_dict = {col: data["input_data"][0][col] for col in input_data_columns}

    results = []
    for backfill_results in data["results_data"]:
        results_dict = {"key": key, **input_data_dict}

        results_dict[outcome_column] = backfill_results[outcome_column]
        results_dict[backfill_id_column] = backfill_results[backfill_id_column]
        results.append(results_dict)

    return [beam.Row(**r) for r in results]


def _make_metrics_transform(
    target: Target,
    group_by_cols: list[str],
) -> beam.PTransform:
    if target.kind == TargetKind.BINARY_DISTRIBUTION:
        return BinaryDistributionTransform(
            target=target,
            group_by_cols=group_by_cols,
        )
    raise NotImplementedError(f"Unsupported target kind for {Target}")


if __name__ == "__main__":
    parser = ArgumentParser()
    # Application-defined
    parser.add_argument("--input_data_path", type=str, help="Path to input data")
    parser.add_argument("--results_data_path", type=str, help="Path to results data")
    parser.add_argument("--output_path", type=str, help="Path to save metrics")
    parser.add_argument("--outcome", type=str, help="Name of the outcome column")
    parser.add_argument(
        "--target_kind", type=str, help="Kind of target, such as BINARY_DISTRIBUTION"
    )
    # User-defined
    parser.add_argument("--target_name", type=str, help="Name of the target column")
    parser.add_argument(
        "--time",
        type=str,
        default=None,
        required=False,
        help="Column that indicates time",
    )
    parser.add_argument(
        "--groups",
        type=str,
        default=None,
        required=False,
        help="Array of columns used to group metrics",
    )

    pipeline_args, unknown_args = parser.parse_known_args()
    target = Target(
        name=pipeline_args.target_name, kind=TargetKind(pipeline_args.target_kind)
    )

    groups = None
    if pipeline_args.groups is not None:
        groups = json.loads(pipeline_args.groups)

    args_pipeline_options = [
        "--save_main_session",
        "--sdk_location",
        "container",
        *unknown_args,
    ]

    metrics_metadata = MetricsMetadata(
        outcome=pipeline_args.outcome,
        target=target,
        time=pipeline_args.time,
        groups=groups,
    )
    pipeline_options = PipelineOptions(args_pipeline_options)

    build_metrics_pipeline(
        input_data_path=pipeline_args.input_data_path,
        results_data_path=pipeline_args.results_data_path,
        output_path=pipeline_args.output_path,
        metrics_metadata=metrics_metadata,
        pipeline_options=pipeline_options,
    ).run().wait_until_finish()



from .base import Node, NodeDefinition, NodeType
from .branch import BranchNode
from .component import ComponentNode
from .connection import ConnectionNode
from .data_input import DataInputNode
from .decision import DecisionNode
from .input import InputNode
from .policy_definition import PolicyDefinitionNode
from .terminate import TerminateNode
from .transform import TransformNode

__all__ = [
    "Node",
    "NodeType",
    "NodeDefinition",
    "InputNode",
    "TransformNode",
    "BranchNode",
    "TerminateNode",
    "DataInputNode",
    "PolicyDefinitionNode",
    "ConnectionNode",
    "DecisionNode",
    "ComponentNode",
]



from typing import Any, Callable, cast

from vulkan.spec.nodes.base import Node, NodeDefinition, NodeType
from vulkan.spec.nodes.metadata import TransformNodeMetadata
from vulkan.spec.nodes.user_code import (
    UserCodeException,
    get_source_code,
    get_udf_instance,
)


class TransformNode(Node):
    """Evaluate an arbitrary function.

    Transform nodes are used to evaluate arbitrary functions.
    They can be thought of as functions, and can be used to transform
    data (hence the name), perform calculations, etc.
    At the moment, there is no limit to the type of functions that can be
    evaluated by a transform node, provided that the function is serializable.
    """

    def __init__(
        self,
        name: str,
        func: Callable,
        dependencies: dict[str, Any],
        description: str | None = None,
        hierarchy: list[str] | None = None,
        parameters: dict[str, str] | None = None,
    ):
        """Evaluate an arbitrary function.

        In the current implementation, the function always receives an
        execution context as its first argument.
        This context can be used for logging via the `ctx.log` attribute.
        Dependencies are passed as keyword arguments.
        keyword arguments, eg.:
        ```
        def add(ctx, a, b):
            ctx.logger.info(f"Adding {a} and {b}")
            return a + b

        node = TransformNode(
            name="add_node",
            func=add,
            dependencies={"a": Dependency("source_a"), "b": Dependency("source_b")},
        )
        ```

        Parameters
        ----------
        name : str
            The name of the node.
        func: callable
            An arbitrary function that will be executed when the node is run.
            The function should receive the dependencies as arguments.
        dependencies: dict, optional
            The dependencies of the node.
            See `Dependency` for more information.
        description: str, optional
            A description of the node.

        """
        super().__init__(
            name=name,
            description=description,
            typ=NodeType.TRANSFORM,
            dependencies=dependencies,
            hierarchy=hierarchy,
        )
        if not callable(func):
            raise TypeError(f"Expected callable, got ({type(func)})")
        self.func = func
        self.parameters = parameters

    def node_definition(self) -> NodeDefinition:
        # Get the text of source code for the user function
        source_code: str = get_source_code(self.func)

        return NodeDefinition(
            name=self.name,
            description=self.description,
            node_type=self.type.value,
            dependencies=self.dependencies,
            metadata=TransformNodeMetadata(
                source_code=source_code,
                parameters=self.parameters,
            ),
            hierarchy=self.hierarchy,
        )

    @classmethod
    def from_dict(cls, spec: dict[str, Any]) -> "TransformNode":
        definition = NodeDefinition.from_dict(spec)
        if definition.metadata is None:
            raise ValueError(f"Metadata not set for node {definition.name}")

        metadata = cast(TransformNodeMetadata, definition.metadata)

        try:
            # Load the stringified source code into a Python executable and
            # encapsulate it in a callable object.
            udf_instance: Callable = get_udf_instance(
                metadata.source_code, definition.dependencies
            )
        except UserCodeException as e:
            raise ValueError(f"Invalid user code in node {definition.name}") from e

        return cls(
            name=definition.name,
            description=definition.description,
            dependencies=definition.dependencies,
            func=udf_instance,
            hierarchy=definition.hierarchy,
            parameters=metadata.parameters,
        )



import apache_beam as beam
from apache_beam.transforms.combiners import CountCombineFn

from vulkan.core.backtest.metrics import Target


class BinaryDistributionTransform(beam.PTransform):
    """Calculate basic statistics for a binary target.

    When the target is binary, and the outcome is a categorical value
    for which the target distribution is to be calculated.
    """

    def __init__(self, target: Target, group_by_cols: list[str]):
        self.target = target
        self.group_by_cols = group_by_cols

    def expand(self, pcoll):
        return pcoll | "Basic Statistics for Binary Distribution" >> (
            beam.GroupBy(*self.group_by_cols)
            .aggregate_field("key", CountCombineFn(), "count")
            .aggregate_field(lambda x: getattr(x, self.target.name) == 1, sum, "ones")
            .aggregate_field(lambda x: getattr(x, self.target.name) == 0, sum, "zeros")
        )

    def output_columns(self):
        return {
            *self.group_by_cols,
            "count",
            "ones",
            "zeros",
        }



import json
import os
from abc import ABC, abstractmethod
from time import time

import pyarrow.parquet as pq
from apache_beam.options.pipeline_options import PipelineOptions

from vulkan.core.policy import Policy
from vulkan.runners.beam.pipeline import LOCAL_RESULTS_FILE_NAME, BeamPipelineBuilder
from vulkan.schemas import DataSourceSpec
from vulkan.spec.policy import PolicyDefinition


class RunResult(ABC):
    @property
    @abstractmethod
    def data(self):
        pass

    @property
    @abstractmethod
    def metadata(self):
        pass

    @abstractmethod
    def _get_output(self):
        pass


class SingleRunResult(RunResult):
    def __init__(self, output_path):
        self.output_path = output_path

    @property
    def data(self):
        return self._get_output()

    @property
    def metadata(self):
        pass

    def _get_output(self):
        output_path = os.path.join(self.output_path, LOCAL_RESULTS_FILE_NAME)
        with open(output_path, "r") as fp:
            results = json.load(fp)

        return results[1]


class BatchRunResult(RunResult):
    def __init__(self, output_path):
        self.output_path = output_path

    @property
    def data(self):
        return self._get_output()

    @property
    def metadata(self):
        pass

    def _get_output(self):
        dataset = pq.ParquetDataset(self.output_path)
        return dataset.read().to_pandas()


class PolicyRunner:
    def __init__(self, policy_definition: PolicyDefinition, staging_path: str):
        self.policy = Policy.from_definition(policy_definition)
        self.policy_definition = policy_definition
        self.staging_path = staging_path

    def run(
        self,
        input_data: dict,
        data_sources: list[DataSourceSpec] | None = None,
        config_variables: dict | None = None,
    ) -> RunResult:
        run_id = str(time()).replace(".", "")
        output_path = os.path.join(self.staging_path, run_id)

        if data_sources is None:
            data_sources = []

        builder = get_pipeline_builder(
            policy=self.policy,
            data_sources=data_sources,
            output_path=output_path,
            config_variables=config_variables,
        )

        options = PipelineOptions(["--runner=DirectRunner"])
        pipeline = builder.build_single_run_pipeline(
            input_data=input_data, pipeline_options=options
        )
        pipeline.run()

        return SingleRunResult(output_path)

    def run_batch(
        self,
        input_data_path: str,
        data_sources: list[DataSourceSpec] | None = None,
        config_variables: dict | None = None,
        run_id: str | None = None,
    ) -> BatchRunResult:
        if run_id is None:
            run_id = str(time()).replace(".", "")
        output_path = os.path.join(self.staging_path, run_id)

        if data_sources is None:
            data_sources = []

        builder = get_pipeline_builder(
            policy=self.policy,
            data_sources=data_sources,
            output_path=output_path,
            config_variables=config_variables,
        )

        options = PipelineOptions(["--runner=DirectRunner"])
        pipeline = builder.build_batch_pipeline(
            backfill_id=run_id,
            input_data_path=input_data_path,
            pipeline_options=options,
        )
        pipeline.run()

        return BatchRunResult(output_path)


def get_pipeline_builder(
    policy,
    data_sources: list[DataSourceSpec],
    output_path: str,
    config_variables: dict[str, str] | None = None,
):
    if config_variables is None:
        config_variables = {}

    data_sources_map = {}
    for source in data_sources:
        assert isinstance(source, DataSourceSpec), (
            f"Data sources should be an instance of DataSourceSpec, got {source}"
        )
        if source.name in data_sources_map:
            raise ValueError(f"Duplicate data source name: {source.name}")
        data_sources_map[source.name] = source

    builder = BeamPipelineBuilder(
        policy=policy,
        output_path=output_path,
        data_sources=data_sources_map,
        config_variables=config_variables,
    )
    return builder



import ast
import re
from typing import Any, Literal

from jinja2 import Environment, Template, nodes
from pydantic import BaseModel, TypeAdapter, ValidationError

_BaseType = str | int | float | bool
_ListType = list[_BaseType]
_ShallowDictType = dict[str, _BaseType | _ListType]
_ValueType = _BaseType | _ListType | _ShallowDictType
_DictType = dict[str, _ValueType]


class EnvVarConfig(BaseModel):
    env: str


class RunTimeParam(BaseModel):
    param: str
    value_type: Literal["str", "int", "float", "auto"] = "auto"


_ParameterType = _BaseType | _ListType | _DictType | EnvVarConfig | RunTimeParam
ConfigurableDict = dict[str, _ParameterType]
configurable_dict_adapter = TypeAdapter(ConfigurableDict)


def _visit_nodes(node, visitor_func):
    """Helper function to recursively visit all nodes in a Jinja2 AST."""
    visitor_func(node)
    for child in getattr(node, "iter_child_nodes", lambda: [])():
        _visit_nodes(child, visitor_func)


def _create_env_var_visitor(vars_list: list):
    """Creates a visitor function to extract environment variables.

    Extracts environment variable names from Jinja2 templates (e.g., {{env.VAR}}).
    Only matches attribute access on the 'env' object (Getattr nodes).
    """

    def visitor(node):
        # Only extract Getattr nodes where the base object is the 'env' Name node
        # This matches patterns like {{env.MY_VAR}}
        if isinstance(node, nodes.Getattr):
            if isinstance(node.node, nodes.Name) and node.node.name == "env":
                vars_list.append(node.attr)

    return visitor


def _create_runtime_param_visitor(vars_list: list):
    """Creates a visitor function to extract runtime parameters.

    Extracts simple variable names from Jinja2 templates (e.g., {{variable}}).
    Excludes environment variables ({{env.VAR}}) and the 'env' object itself.

    This visitor handles RunTimeParam instances which may have different value_type
    specifications (str, int, float, auto), but at the template level we only
    extract the variable names - type information is preserved in the RunTimeParam
    object itself, not in the template string.
    """

    def visitor(node):
        if isinstance(node, nodes.Name) and node.name != "env":
            vars_list.append(node.name)

    return visitor


def _extract_variables_from_template_string(
    template_string: str, visitor_factory
) -> list[str]:
    """
    Helper to extract variables from a Jinja2 template string using a provided visitor factory.
    """
    if not template_string or not isinstance(template_string, str):
        return []

    extracted_vars = []
    env = Environment()

    ast = env.parse(template_string)
    visitor_func = visitor_factory(extracted_vars)
    _visit_nodes(ast, visitor_func)

    return extracted_vars


def extract_env_vars(spec: ConfigurableDict) -> list[str]:
    """
    Extract environment variable names from a ConfigurableDict using Jinja2 AST parsing.

    Returns a list of environment variable names found in the spec.
    Raises TemplateSyntaxError for invalid Jinja2 syntax.
    """
    if spec is None:
        return []

    env_vars = []

    for key, value in spec.items():
        # Try to convert dict to RunTimeParam or EnvVarConfig if applicable
        value = _try_cast_to_config_object(value)

        if isinstance(value, EnvVarConfig):
            env_vars.append(value.env)
        elif isinstance(value, str):
            try:
                env_vars.extend(
                    _extract_variables_from_template_string(
                        value, _create_env_var_visitor
                    )
                )
            except Exception as e:
                raise ValueError(
                    f"Invalid Jinja2 template in field '{key}': {value}"
                ) from e

    return list(set(env_vars))


def extract_runtime_params(spec: ConfigurableDict) -> list[str]:
    """
    Extract runtime parameter names from a ConfigurableDict using Jinja2 AST parsing.

    Returns a list of runtime parameter names found in the spec.
    Raises TemplateSyntaxError for invalid Jinja2 syntax.
    """
    if spec is None:
        return []

    runtime_params = []

    for key, value in spec.items():
        # Try to convert dict to RunTimeParam or EnvVarConfig if applicable
        value = _try_cast_to_config_object(value)

        if isinstance(value, RunTimeParam):
            runtime_params.append(value.param)
        elif isinstance(value, str):
            try:
                runtime_params.extend(
                    _extract_variables_from_template_string(
                        value, _create_runtime_param_visitor
                    )
                )
            except Exception as e:
                raise ValueError(
                    f"Invalid Jinja2 template in field '{key}': {value}"
                ) from e

    return list(set(runtime_params))


def extract_env_vars_from_string(template: str) -> list[str]:
    """
    Extract environment variable names from a template using Jinja2 AST parsing.

    Returns a list of environment variable names found in the template.
    Raises ValueError for invalid Jinja2 syntax.
    """
    if not template or not isinstance(template, str):
        return []

    try:
        return list(
            set(
                _extract_variables_from_template_string(
                    template, _create_env_var_visitor
                )
            )
        )
    except Exception as e:
        raise ValueError(f"Invalid Jinja2 template: {template}") from e


def extract_runtime_params_from_string(template: str) -> list[str]:
    """
    Extract runtime parameter names from a template using Jinja2 AST parsing.

    Returns a list of runtime parameter names found in the template.
    Raises ValueError for invalid Jinja2 syntax.
    """
    if not template or not isinstance(template, str):
        return []

    try:
        return list(
            set(
                _extract_variables_from_template_string(
                    template, _create_runtime_param_visitor
                )
            )
        )
    except Exception as e:
        raise ValueError(f"Invalid Jinja2 template: {template}") from e


def configure_fields(
    spec: ConfigurableDict, local_variables: dict, env_variables: dict
) -> dict:
    """
    Configure fields in a specification dictionary by resolving templates.

    This function replaces any RunTimeParam or EnvVarConfig instances with their
    corresponding template strings, and resolves any string templates using
    the provided node and environment variables.

    Dictionary casting:
    - Dicts with "param" key are automatically converted to RunTimeParam
    - Dicts with "env" key are automatically converted to EnvVarConfig
    - This enables seamless handling of both object and serialized representations

    Type handling:
    - RunTimeParam: Uses the value_type field to determine expected type
    - EnvVarConfig: Always returns strings (matching OS environment variable semantics)
    - Plain templates: Uses "auto" type inference
    """
    if spec is None:
        spec = {}

    for key, value in spec.items():
        expected_type = "auto"

        # Try to convert dict to RunTimeParam or EnvVarConfig if applicable
        value = _try_cast_to_config_object(value)

        if isinstance(value, RunTimeParam):
            expected_type = value.value_type
            value = normalize_to_template(value)
        elif isinstance(value, EnvVarConfig):
            # Environment variables are always strings (OS semantics)
            expected_type = "str"
            value = normalize_to_template(value)

        if _is_template_like(value):
            spec[key] = resolve_template(
                value, local_variables, env_variables, expected_type
            )

    return spec


def _try_cast_to_config_object(value: Any) -> RunTimeParam | EnvVarConfig | Any:
    """
    Try to convert a dict to RunTimeParam or EnvVarConfig if applicable.

    Args:
        value: The value to potentially convert

    Returns:
        RunTimeParam if dict has "param" key and is valid
        EnvVarConfig if dict has "env" key and is valid
        Original value otherwise
    """
    if not isinstance(value, dict):
        return value

    # Try to convert to RunTimeParam
    if "param" in value:
        try:
            return RunTimeParam(**value)
        except ValidationError:
            pass

    # Try to convert to EnvVarConfig
    if "env" in value:
        try:
            return EnvVarConfig(**value)
        except ValidationError:
            pass

    return value


def _is_template_like(value: Any) -> bool:
    if not isinstance(value, str):
        return False

    # Check if the string contains at least one Jinja2 template expression
    return bool(re.search(r"\{\{.*?\}\}", value))


def _convert_to_type(value: str, target_type: str) -> int | float | bool:
    """
    Convert a string value to the specified type with clear error messages.

    Args:
        value: The string value to convert
        target_type: One of "int", "float"

    Returns:
        The converted value

    Raises:
        ValueError: If the conversion fails
    """
    if target_type not in ("int", "float"):
        raise ValueError(f"Unsupported target type for conversion: {target_type}")

    try:
        if target_type == "int":
            return int(value)
        elif target_type == "float":
            return float(value)
    except (ValueError, TypeError, AttributeError) as e:
        raise ValueError(
            f"Cannot convert template result '{value}' to {target_type}: {e}"
        )


def resolve_template(
    value: str,
    local_variables: dict,
    env_variables: dict,
    expected_type: str = "auto",
) -> _ValueType:
    """
    Resolve a Jinja2 template string using provided local and environment variables.

    Args:
        value: The template string to resolve (e.g., "{{variable}}")
        local_variables: Dictionary of runtime variables
        env_variables: Dictionary of environment variables
        expected_type: Expected type of the result. One of:
            - "str": Always return a string (no type conversion)
            - "int": Convert to integer
            - "float": Convert to float
            - "auto": Attempt to infer type using ast.literal_eval (default)

    Returns:
        The resolved value with the appropriate type

    Raises:
        ValueError: If template is invalid or type conversion fails
    """
    if not isinstance(value, str):
        raise ValueError(
            f"Expected a string for template resolution, got {type(value)}"
        )

    if not value.strip():
        return value

    template = Template(value)
    context = {"env": env_variables, **local_variables}
    rendered = template.render(context)

    # Explicit string type - no conversion needed
    if expected_type == "str":
        return rendered

    # Explicit typed conversion
    if expected_type in ("int", "float"):
        return _convert_to_type(rendered, expected_type)

    # Auto - infer type using ast.literal_eval
    # This attempts to evaluate the rendered template as a Python literal.
    # If it looks like a bool, list, dict, etc, it will be converted.
    try:
        return ast.literal_eval(rendered)
    except (ValueError, SyntaxError):
        return rendered


def resolve_value(
    value: Any,
    jinja_context: dict[str, Any],
    env_variables: dict[str, Any],
) -> Any:
    """
    Recursively resolve values that may contain templates.

    Handles strings (templates), dictionaries, and lists by recursively
    resolving any template expressions found within them.

    Args:
        value: The value to resolve (str, dict, list, or other)
        jinja_context: Dictionary of local/runtime variables
        env_variables: Dictionary of environment variables

    Returns:
        The resolved value with templates replaced

    Examples:
        >>> resolve_value("{{x}}", {"x": 5}, {})
        5
        >>> resolve_value({"a": "{{x}}"}, {"x": 5}, {})
        {"a": 5}
        >>> resolve_value(["{{x}}", "{{y}}"], {"x": 5, "y": 10}, {})
        [5, 10]
    """
    if isinstance(value, str):
        try:
            # Use Jinja2-based template resolution with environment variables
            return resolve_template(value, jinja_context, env_variables)
        except Exception:
            # If template resolution fails, return original value
            return value
    elif isinstance(value, dict):
        return {
            k: resolve_value(v, jinja_context, env_variables) for k, v in value.items()
        }
    elif isinstance(value, list):
        return [resolve_value(item, jinja_context, env_variables) for item in value]
    else:
        return value


def normalize_to_template(value: _ParameterType) -> str:
    """
    Normalize both object-based and template-based configurations to template format.

    Examples:
    - {"param": "variable"} -> "{{variable}}"
    - {"env": "variable"} -> "{{env.variable}}"
    - "{{variable}}" -> "{{variable}}" (unchanged)
    - "variable" -> "{{variable}}"
    """
    if isinstance(value, RunTimeParam):
        return f"{{{{{value.param}}}}}"
    elif isinstance(value, EnvVarConfig):
        return f"{{{{env.{value.env}}}}}"
    elif isinstance(value, list):
        # For lists, normalize each element
        return [normalize_to_template(item) for item in value]
    elif isinstance(value, str) and value.startswith("{{") and value.endswith("}}"):
        return value
    else:
        return f"{{{{{value}}}}}"


def normalize_mapping(spec: ConfigurableDict) -> dict[str, str]:
    """
    Normalize a ConfigurableDict to template format.

    Returns a dictionary where all values are normalized to template strings.
    """
    if spec is None:
        return {}

    return {key: normalize_to_template(value) for key, value in spec.items()}



import csv
import hashlib
import json
from functools import partial

import apache_beam as beam
import pyarrow as pa
from apache_beam.dataframe import convert
from apache_beam.dataframe.io import read_csv


class ReadParquet(beam.PTransform):
    def __init__(self, data_path: str):
        self.data_path = data_path

    def expand(self, pcoll):
        return (
            pcoll
            | "Read File" >> beam.io.ReadFromParquet(self.data_path)
            | "Make Key" >> beam.Map(_make_element_key)
        )


class ReadRemoteCSV(beam.PTransform):
    def __init__(self, data_path: str, schema: dict[str, type] | None = None):
        self.data_path = data_path
        self.csv_parser = partial(parse_csv_line, schema)

    def expand(self, pcoll):
        return (
            pcoll
            | "Read File" >> beam.io.ReadFromText(self.data_path, skip_header_lines=1)
            | "Parse CSV" >> beam.Map(self.csv_parser)
            | "Make Key" >> beam.Map(_make_element_key)
        )


def parse_csv_line(schema, line):
    """Parse a single line of CSV into a dictionary."""
    reader = csv.reader([line])
    fields = list(schema.keys())
    data = dict(zip(fields, next(reader)))
    return {k: schema[k](v) for k, v in data.items()}


class ReadLocalCSV(beam.PTransform):
    def __init__(self, source: str):
        self.source = source

    def expand(self, pcoll):
        dataframe = pcoll | "Read Source" >> read_csv(self.source)
        return (
            convert.to_pcollection(dataframe)
            | "To dictionaries" >> beam.Map(lambda x: dict(x._asdict()))
            | "Make Key" >> beam.Map(_make_element_key)
        )


def _make_element_key(element: dict) -> tuple[str, dict]:
    content = json.dumps(element, sort_keys=True)
    key = hashlib.md5(content.encode("utf-8")).hexdigest()
    return (key, element)


class ProcessResultFn(beam.DoFn):
    def __init__(self, backfill_id: str) -> None:
        self.backfill_id = backfill_id

    def process(self, element, **kwargs):
        result = {}
        key, value = element
        result["backfill_id"] = self.backfill_id
        result["key"] = key
        result["status"] = value["result"][0]["status"]
        result.update({k: v[0] for k, v in value.items() if k != "result"})
        yield result


class WriteParquet(beam.PTransform):
    def __init__(self, filepath: str, schema: dict[str, type], backfill_id: str):
        self.filepath = filepath
        self.schema = to_pyarrow_schema(schema)
        self.backfill_id = backfill_id

    def expand(self, pcoll):
        return (
            pcoll
            | "Process Result" >> beam.ParDo(ProcessResultFn(self.backfill_id))
            | "Write to Parquet" >> beam.io.WriteToParquet(self.filepath, self.schema)
        )


def to_pyarrow_schema(schema: dict[str, type]) -> dict[str, str]:
    fields = [to_pyarrow_field(name, dtype) for name, dtype in schema.items()]
    return pa.schema(fields)


def to_pyarrow_field(name, dtype):
    if isinstance(dtype, dict):
        # Handle nested structure (dict) by recursively converting fields
        nested_fields = [to_pyarrow_field(k, v) for k, v in dtype.items()]
        return pa.field(name, pa.struct(nested_fields))

    if isinstance(dtype, list):
        # Handle list types (e.g., list of strings).
        # Currently assuming a list of a single primitive type.
        # TODO: add support for more complex nested structures.
        element_type = _PYARROW_TYPE_MAPPING[dtype[0]]
        return pa.field(name, pa.list(element_type))

    # Directly map primitive types
    return pa.field(name, _PYARROW_TYPE_MAPPING[dtype])


_PYARROW_TYPE_MAPPING = {
    int: pa.int64(),
    float: pa.float64(),
    str: pa.string(),
    bool: pa.bool_(),
}



from dagster import ConfigurableResource


class VulkanRunConfig(ConfigurableResource):
    run_id: str
    server_url: str
    project_id: str | None = None


class VulkanPolicyConfig(ConfigurableResource):
    variables: dict



from abc import ABC, abstractmethod
from enum import Enum

from pydantic import BaseModel, field_validator

from vulkan.auth import AuthConfig
from vulkan.connections import HTTPConfig
from vulkan.credentials import validate_no_reserved_credentials_in_templates
from vulkan.node_config import (
    extract_env_vars,
    extract_env_vars_from_string,
    extract_runtime_params,
    extract_runtime_params_from_string,
)


class DataSourceType(Enum):
    HTTP = "http"
    LOCAL_FILE = "local_file"
    REGISTERED_FILE = "registered_file"


class DataSourceStatus(Enum):
    DRAFT = "DRAFT"
    PUBLISHED = "PUBLISHED"
    ARCHIVED = "ARCHIVED"


class SourceSpecBase(ABC):
    @abstractmethod
    def extract_env_vars(self) -> list[str]:
        pass

    @abstractmethod
    def extract_runtime_params(self) -> list[str]:
        pass

    @property
    @abstractmethod
    def source_type(self) -> DataSourceType:
        pass


class RegisteredFileSource(BaseModel, SourceSpecBase):
    file_id: str
    path: str | None = None

    def extract_env_vars(self) -> list[str]:
        return []

    def extract_runtime_params(self) -> list[str]:
        return []

    @property
    def source_type(self) -> DataSourceType:
        return DataSourceType.REGISTERED_FILE


class LocalFileSource(BaseModel):
    path: str

    def extract_env_vars(self) -> list[str]:
        return []

    def extract_runtime_params(self) -> list[str]:
        return []

    @property
    def source_type(self) -> DataSourceType:
        return DataSourceType.LOCAL_FILE


class HTTPSource(HTTPConfig, SourceSpecBase):
    """HTTP data source configuration with optional authentication."""

    auth: AuthConfig | None = None

    def extract_env_vars(self) -> list[str]:
        """
        Extracts environment variables from URL, headers, params, and body.

        Note: Does NOT include CLIENT_ID/CLIENT_SECRET.
        Those are reserved for authentication and managed separately.
        """
        env_vars = extract_env_vars_from_string(self.url)
        for spec in [self.headers, self.params, self.body]:
            if spec is not None:
                env_vars += extract_env_vars(spec)
        return env_vars

    def extract_runtime_params(self) -> list[str]:
        """Extracts runtime parameters from URL, headers, params, and body."""
        params = extract_runtime_params_from_string(self.url)
        for spec in [self.headers, self.params, self.body]:
            if spec is not None:
                params += extract_runtime_params(spec)
        return params

    @field_validator("auth", mode="after")
    @classmethod
    def validate_auth_secrets_not_in_templates(cls, v, info):
        """
        Validates that reserved credential names are not used in templates.

        Uses shared validation function to ensure consistency across the codebase.
        """
        if v is None:
            return v

        data = info.data

        env_vars = []
        if "url" in data:
            env_vars += extract_env_vars_from_string(data["url"])
        for field in ["headers", "params", "body"]:
            if field in data and data[field] is not None:
                env_vars += extract_env_vars(data[field])

        validate_no_reserved_credentials_in_templates(
            env_vars, error_context="templates"
        )

        return v

    @property
    def source_type(self) -> DataSourceType:
        return DataSourceType.HTTP



from unittest.mock import MagicMock

from hatchet_sdk import Hatchet
from vulkan.core.policy import Policy
from vulkan.runners.hatchet.nodes import (
    HatchetDataInput,
    HatchetTerminate,
    HatchetTransform,
    to_hatchet_nodes,
)
from vulkan.runners.hatchet.policy import HatchetFlow
from vulkan.spec.dependency import INPUT_NODE, Dependency
from vulkan.spec.nodes import BranchNode, DataInputNode, TerminateNode, TransformNode
from vulkan.spec.policy import PolicyDefinition


def test_data_input_node_conversion():
    """Test DataInputNode to HatchetDataInput conversion."""
    node = DataInputNode(
        name="test_data_input",
        data_source="test_source",
        description="Test data input",
        parameters={"param1": "value1"},
    )

    hatchet_node = HatchetDataInput.from_spec(node)

    assert hatchet_node.name == "test_data_input"
    assert hatchet_node.data_source == "test_source"
    assert hatchet_node.description == "Test data input"
    assert hatchet_node.parameters == {"param1": "value1"}
    assert hatchet_node.task_name == "test_data_input"
    assert callable(hatchet_node.task_fn())


def test_transform_node_conversion():
    """Test TransformNode to HatchetTransform conversion."""

    def test_func(x):
        return x * 2

    node = TransformNode(
        name="test_transform",
        description="Test transform",
        func=test_func,
        dependencies={"input": Dependency(INPUT_NODE)},
    )

    hatchet_node = HatchetTransform.from_spec(node)

    assert hatchet_node.name == "test_transform"
    assert hatchet_node.description == "Test transform"
    assert hatchet_node.func == test_func
    assert hatchet_node.task_name == "test_transform"
    assert callable(hatchet_node.task_fn())


def test_terminate_node_conversion():
    """Test TerminateNode to HatchetTerminate conversion."""
    node = TerminateNode(
        name="test_terminate",
        description="Test terminate",
        return_status="SUCCESS",
        dependencies={"input": Dependency(INPUT_NODE)},
    )

    hatchet_node = HatchetTerminate.from_spec(node)

    assert hatchet_node.name == "test_terminate"
    assert hatchet_node.description == "Test terminate"
    assert hatchet_node.return_status == "SUCCESS"
    assert hatchet_node.task_name == "test_terminate"
    assert callable(hatchet_node.task_fn())


def test_to_hatchet_nodes():
    """Test conversion of multiple nodes."""
    nodes = [
        DataInputNode(name="data_input", data_source="test", dependencies={}),
        TransformNode(
            name="transform", description="test", func=lambda x: x, dependencies={}
        ),
        TerminateNode(
            name="terminate",
            description="test",
            return_status="SUCCESS",
            dependencies={},
        ),
    ]

    hatchet_nodes = to_hatchet_nodes(nodes)

    assert len(hatchet_nodes) == 3
    assert isinstance(hatchet_nodes[0], HatchetDataInput)
    assert isinstance(hatchet_nodes[1], HatchetTransform)
    assert isinstance(hatchet_nodes[2], HatchetTerminate)


def test_regression_hatchet_flow_with_unsorted_node_list():
    def is_even_fn(input_node):
        if input_node["number"] % 2 == 0:
            return "even"
        return "odd"

    branch_node_1 = BranchNode(
        name="branch_node_1",
        func=is_even_fn,
        choices=["even", "odd"],
        dependencies={INPUT_NODE: Dependency(INPUT_NODE)},
    )

    is_even = TerminateNode(
        name="is_even",
        return_status="Great",
        dependencies={"condition": Dependency("branch_node_1", "even")},
    )

    is_odd = TerminateNode(
        name="is_odd",
        return_status="Too_Bad",
        dependencies={"condition": Dependency("branch_node_1", "odd")},
    )

    policy_def = PolicyDefinition(
        nodes=[
            is_even,
            is_odd,
            branch_node_1,
        ],
        input_schema={"number": "int"},
    )
    policy = Policy.from_definition(policy_def)

    flow = HatchetFlow(
        nodes=policy.nodes,
        hatchet=Hatchet(client=MagicMock(), config=MagicMock()),
    )
    flow.create_workflow()



from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Any, Literal

from pydantic import BaseModel

from vulkan.spec.dependency import Dependency, DependencyDict
from vulkan.spec.nodes.metadata import (
    BaseNodeMetadata,
    BranchNodeMetadata,
    ComponentNodeMetadata,
    ConnectionNodeMetadata,
    DataInputNodeMetadata,
    DecisionNodeMetadata,
    InputNodeMetadata,
    PolicyNodeMetadata,
    TerminateNodeMetadata,
    TransformNodeMetadata,
)


class NodeType(Enum):
    TRANSFORM = "TRANSFORM"
    TERMINATE = "TERMINATE"
    BRANCH = "BRANCH"
    INPUT = "INPUT"
    DATA_INPUT = "DATA_INPUT"
    POLICY = "POLICY"
    CONNECTION = "CONNECTION"
    DECISION = "DECISION"
    COMPONENT = "COMPONENT"


NODE_METADATA_TYPE_MAP = {
    NodeType.TRANSFORM: TransformNodeMetadata,
    NodeType.TERMINATE: TerminateNodeMetadata,
    NodeType.BRANCH: BranchNodeMetadata,
    NodeType.INPUT: InputNodeMetadata,
    NodeType.DATA_INPUT: DataInputNodeMetadata,
    NodeType.POLICY: PolicyNodeMetadata,
    NodeType.CONNECTION: ConnectionNodeMetadata,
    NodeType.DECISION: DecisionNodeMetadata,
    NodeType.COMPONENT: ComponentNodeMetadata,
}


class NodeDefinitionDict(BaseModel):
    """Dict representation of a NodeDefinition object."""

    name: str
    node_type: str
    dependencies: dict[str, DependencyDict] | None = None
    metadata: dict | None = None
    description: str | None = None
    hierarchy: list[str] | None = None


@dataclass()
class NodeDefinition:
    "Internal representation of a node."

    name: str
    node_type: str
    description: str | None = None
    dependencies: dict[str, DependencyDict] | None = None
    metadata: BaseNodeMetadata | None = None
    hierarchy: list[str] | None = None

    _REQUIRED_KEYS = {"name", "node_type"}

    def __post_init__(self):
        if self.metadata is not None:
            if not isinstance(self.metadata, BaseNodeMetadata):
                msg = f"Metadata must be an instance of ComponentNodeMetadata, got {type(self.metadata)}"
                raise TypeError(msg)
        if self.dependencies is not None:
            deps = {}
            for key, dep in self.dependencies.items():
                if isinstance(dep, Dependency):
                    deps[key] = dep
                else:
                    try:
                        dep = Dependency.from_dict(dep)
                        deps[key] = dep
                    except Exception as e:
                        msg = f"Error parsing dependency {key}: {e}"
                        raise ValueError(msg) from e
            self.dependencies = deps
            if not all(isinstance(d, Dependency) for d in self.dependencies.values()):
                msg = f"Dependencies must be of type Dependency: {self.dependencies}"
                raise ValueError(msg)

    @classmethod
    def from_dict(cls, data: NodeDefinitionDict) -> "NodeDefinition":
        missing_keys = cls._REQUIRED_KEYS - set(data.keys())
        if missing_keys:
            raise ValueError(f"Missing keys: {missing_keys}")

        metadata = data.get("metadata", None)
        if metadata is not None:
            node_type = NodeType(data["node_type"])
            metadata = NODE_METADATA_TYPE_MAP[node_type].from_dict(metadata)

        return cls(
            name=data["name"],
            node_type=data["node_type"],
            description=data.get("description"),
            dependencies=data.get("dependencies"),
            metadata=metadata,
            hierarchy=data.get("hierarchy"),
        )

    def to_dict(
        self, mode: str | Literal["json", "python"] = "python"
    ) -> NodeDefinitionDict:
        data: dict[str, Any] = {
            "name": self.name,
            "node_type": self.node_type,
        }
        if self.description is not None:
            data["description"] = self.description
        if self.dependencies is not None:
            data["dependencies"] = {
                key: d.to_dict() for key, d in self.dependencies.items()
            }
        if self.metadata is not None:
            data["metadata"] = self.metadata.to_dict(mode=mode)
        if self.hierarchy is not None:
            data["hierarchy"] = self.hierarchy
        return data


class Node(ABC):
    """A node represents a step in a workflow.

    It can be thought of as a function that executes in an isolated environment.
    Each node represents a vertice in a DAG, and declares its dependencies,
    which are the edges in the graph.
    """

    def __init__(
        self,
        name: str,
        typ: NodeType,
        description: str | None = None,
        dependencies: dict[str, Dependency] | None = None,
        hierarchy: list[str] | None = None,
    ):
        """A node. This is an abstract class and should not be instantiated directly.

        Parameters
        ----------
        name : str
            The name of the node.
        typ : NodeType
            The type of the node. Determines the overall behavior of the node.
        description : str, optional
            A description of the node. Used for documentation purposes and
            shown in the user interface.
        dependencies : dict[str, Dependency], optional
            The dependencies of the node. A dictionary where the key is the name
            of the variable that will receive the data, and the value is the source.
            See `Dependency` for more information.

        """
        self._name = name
        self.description = description
        self.type = typ
        self._dependencies = dependencies if dependencies is not None else {}
        if not isinstance(self._dependencies, dict):
            raise TypeError(f"Dependencies must be a dict, got: {dependencies}")
        if not all(isinstance(d, Dependency) for d in self._dependencies.values()):
            raise TypeError("Dependencies must be of type Dependency")

        self._hierarchy = hierarchy

    @abstractmethod
    def node_definition(self) -> NodeDefinition:
        pass

    @abstractmethod
    def from_dict(cls, spec: NodeDefinitionDict) -> "Node":
        pass

    @property
    def name(self) -> str:
        return self._name

    @property
    def dependencies(self) -> dict[str, Dependency]:
        return self._dependencies

    @property
    def hierarchy(self) -> list[str] | None:
        return self._hierarchy

    @hierarchy.setter
    def hierarchy(self, hierarchy: list[str]):
        self._hierarchy = hierarchy

    @property
    def id(self) -> str:
        if self._hierarchy is None or len(self._hierarchy) == 0:
            return self.name
        return "-".join(self._hierarchy) + "." + self._name

    def node_dependencies(self) -> list[Dependency]:
        return list(self.dependencies.values())

    def to_dict(
        self, mode: str | Literal["json", "python"] = "python"
    ) -> NodeDefinitionDict:
        return self.node_definition().to_dict(mode)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Node):
            return False
        return self.node_definition() == other.node_definition()


class PolicyDefinitionDict(BaseModel):
    """Dict representation of a PolicyDefinition object."""

    nodes: list[NodeDefinitionDict]
    input_schema: dict[str, str]
    config_variables: list[str] | None = None
    output_callback: str | None = None



import builtins
from dataclasses import dataclass, field
from typing import Callable, Literal

from vulkan.spec.dependency import INPUT_NODE
from vulkan.spec.graph import GraphDefinition
from vulkan.spec.nodes import (
    BranchNode,
    ComponentNode,
    ConnectionNode,
    DataInputNode,
    DecisionNode,
    InputNode,
    PolicyDefinitionNode,
    TerminateNode,
    TransformNode,
)
from vulkan.spec.nodes.base import Node, NodeType, PolicyDefinitionDict


@dataclass
class PolicyDefinition(GraphDefinition):
    """A policy definition specifies the workflow of a policy.

    A policy is composed of nodes  and the dependencies between them.
    Each node represents a step in the workflow, and can be
    thought of as a function that executes in an isolated environment.
    It receives data from its dependencies, and produces data that can be
    passed to other nodes.

    Dependencies are the main method to pass data between nodes.
    When a node depends on another, the output of the dependency is used
    as input to the node.

    Dependencies are specified per node, as a dictionary, where the key
    is the name of a variable that will receive the data, and the value
    specifies the source node.
    An example of a dependency is:
    `{"source_node_outputs": Dependency("source_node")}`
    In this case, the parameter `source_node_outputs` will receive the output
    of the `source_node` node.

    Parameters
    ----------
    nodes : list[Node]
        The nodes that compose the policy.
        Each node represents a step in the workflow, and can be thought of as
        a function that executes in an isolated environment.
    input_schema : dict[str, type]
        The input schema of the policy.
        It is a dictionary where the key is the name of the input parameter, and
        the value is the type of the parameter.
    output_callback : Callable, optional
        A callback that is called when the policy finishes execution.
        The callback receives the output of the policy as input.
    config_variables : list[str], optional
        The configuration variables that are used to parameterize policy.
        They provide a way to customize the behavior of the policy without
        changing the underlying logic.

    """

    nodes: list[Node]
    input_schema: dict[str, str]
    output_callback: Callable | None = None
    config_variables: list[str] = field(default_factory=list)
    hierarchy: list[str] = field(default_factory=list)

    def __post_init__(self):
        super().__post_init__()
        self._input_node = make_input_node(self.input_schema, self.hierarchy)

        for node in self.nodes:
            node.hierarchy = self.hierarchy
            for dep in node.dependencies.values():
                dep.hierarchy = self.hierarchy

        if self.output_callback is not None:
            if not callable(self.output_callback):
                raise ValueError("Output callback must be a callable")

        if self.config_variables:
            if not isinstance(self.config_variables, list) or not (
                all(isinstance(i, str) for i in self.config_variables)
            ):
                raise ValueError("config_variables must be a list of strings")

        for node in self.nodes:
            if node.name == INPUT_NODE:
                raise ValueError(f"Node name`{INPUT_NODE}` is reserved")

    def to_dict(
        self, mode: str | Literal["json", "python"] = "python"
    ) -> PolicyDefinitionDict:
        return {
            "nodes": [node.to_dict(mode) for node in self.nodes],
            "input_schema": self.input_schema,
            "output_callback": self.output_callback,
            "config_variables": self.config_variables,
        }

    @classmethod
    def from_dict(
        self,
        spec: PolicyDefinitionDict,
        hierarchy: list[str] | None = None,
    ) -> "PolicyDefinition":
        nodes = [node_from_spec(node) for node in spec["nodes"]]

        return PolicyDefinition(
            nodes=nodes,
            input_schema=spec["input_schema"],
            output_callback=spec.get("output_callback", None),
            config_variables=spec.get("config_variables", []),
            hierarchy=hierarchy,
        )

    @property
    def input_node(self) -> InputNode:
        return self._input_node

    @input_node.setter
    def input_node(self, input_node: InputNode):
        self._input_node = input_node


NODE_IMPLEMENTS = {
    NodeType.TRANSFORM: TransformNode,
    NodeType.TERMINATE: TerminateNode,
    NodeType.INPUT: InputNode,
    NodeType.DATA_INPUT: DataInputNode,
    NodeType.BRANCH: BranchNode,
    NodeType.POLICY: PolicyDefinitionNode,
    NodeType.CONNECTION: ConnectionNode,
    NodeType.DECISION: DecisionNode,
    NodeType.COMPONENT: ComponentNode,
}


def node_from_spec(spec: dict[str, str]) -> Node:
    """Create a node from a specification dictionary.

    Parameters
    ----------
    spec : dict[str, str]
        The specification dictionary containing the node type and other parameters.

    Returns
    -------
    Node
        The created node instance.

    Raises
    ------
    ValueError
        If the node type is not recognized or if the specification is invalid.
    """
    node_type = spec.get("node_type")
    if node_type is None:
        raise ValueError("Missing node_type")

    node_type = NodeType(node_type)

    if node_type not in NODE_IMPLEMENTS:
        raise ValueError(f"Unknown node type: {node_type}")

    return NODE_IMPLEMENTS[node_type].from_dict(spec)


def make_input_node(
    input_schema: dict[str, type | str],
    hierarchy: list[str] | None = None,
) -> InputNode:
    input_schema = _parse_input_schema(input_schema)
    return InputNode(
        name=INPUT_NODE,
        description="Input node",
        schema=input_schema,
        hierarchy=hierarchy,
    )


def _parse_input_schema(input_schema: dict[str, type | str]) -> dict[str, type | str]:
    """Parse the input schema to ensure that all types are valid."""
    parsed_schema = {}
    for key, value in input_schema.items():
        if isinstance(value, str):
            # If the type is a string, we assume it's a built-in type
            if hasattr(builtins, value):
                parsed_schema[key] = getattr(builtins, value)
            else:
                msg = f"Invalid type '{value}' for key '{key}'"
                raise ValueError(msg)
        elif isinstance(value, type):
            parsed_schema[key] = value
        else:
            msg = f"Invalid type for key '{key}': {type(value)}"
            raise ValueError(msg)
    return parsed_schema



from dataclasses import dataclass
from enum import Enum

VULKAN_CONFIG_FILENAME = "vulkan.yaml"


class PackagingMode(Enum):
    PYTHON_MODULE = "python_module"
    PYTHON_PACKAGE = "python_package"


@dataclass
class PackagingConfig:
    mode: PackagingMode
    entrypoint: str | None

    def __post_init__(self):
        if self.mode == PackagingMode.PYTHON_PACKAGE and self.entrypoint is None:
            raise ValueError("entrypoint must be provided for python_module packaging")


@dataclass
class UserWorkspaceConfig:
    packaging: PackagingConfig

    @classmethod
    def from_dict(cls, data: dict) -> "UserWorkspaceConfig":
        return cls(
            packaging=PackagingConfig(
                mode=PackagingMode(data["packaging"]["mode"]),
                entrypoint=data["packaging"].get("entrypoint"),
            ),
        )



from dataclasses import dataclass, field
from typing import Any, Dict


@dataclass
class VulkanRunConfig:
    """Configuration for Vulkan run execution."""

    run_id: str
    server_url: str
    project_id: str | None = None


@dataclass
class VulkanPolicyConfig:
    """Configuration for Vulkan policy execution."""

    variables: Dict[str, Any] = field(default_factory=dict)



from abc import ABC, abstractmethod


class ExecutionContext(ABC):
    """Abstract class for the execution context of a Policy."""

    @property
    @abstractmethod
    def log(self):
        """The logger for the current execution."""

    @property
    @abstractmethod
    def env(self):
        """The environment dictionary for the current execution."""



from dataclasses import dataclass
from typing import TypedDict


class DependencyDict(TypedDict):
    """Dict representation of a Dependency object."""

    node: str
    output: str | None = None
    key: str | None = None
    hierarchy: list[str] | None = None


@dataclass
class Dependency:
    """Allows the user to qualify the dependency between two nodes.

    A dependency may be defined in three ways:
    1. When a single string is passed, it is assumed that the dependency is
         the entire output of the node with the same name.
    2. When node and output are specified, we interpret that the dependency
        is on the `output` product of `node`. This is currently only used
        in branching nodes, where the user can specify which output to use.
        That is, if a branch has two possible outputs (eg left and right),
        the user can specify which one to use as
        `Dependency(node="branch", output="left")`.
    3. When key is specified, the dependency is on the `key` entry of the
        given node. This allows the user to access an internal dictionary
        entry of the node, which can simplify data retrieval.
        Note that this will break if the node does not return a dictionary.

    An example of each scenario:

    ```
    x: Dependency("node")  # Depends on the entire output of "node"
    x: Dependency(node="node", output="condition")  # Depends on "node.condition"
    x: Dependency(node="node", key="key")  # Depends on "node['key']"
    ```

    """

    node: str
    output: str | None = None
    key: str | None = None
    hierarchy: list[str] | None = None

    def __post_init__(self):
        assert isinstance(self.node, str) and len(self.node) > 0, (
            "Node name must be a non-empty string"
        )

        assert self.output is None or self.key is None, (
            "Cannot specify both output and key at the moment"
        )

    @property
    def id(self) -> str:
        """The ID of the dependency.

        This is the full path to the dependency, including the hierarchy.
        """
        hierarchy = "-".join(self.hierarchy) if self.hierarchy is not None else ""

        str_repr = self.node
        if len(hierarchy) > 0:
            str_repr = f"{hierarchy}.{self.node}"

        return str_repr

    def __str__(self) -> str:
        node_id = self.id

        if self.output is not None:
            return node_id + f".{self.output}"
        if self.key is not None:
            return node_id + f"[{self.key}]"
        return node_id

    @classmethod
    def from_dict(cls, data: DependencyDict) -> "Dependency":
        return cls(
            node=data["node"],
            output=data["output"],
            key=data["key"],
            hierarchy=data.get("hierarchy"),
        )

    def to_dict(self) -> DependencyDict:
        return DependencyDict(
            node=self.node,
            output=self.output,
            key=self.key,
            hierarchy=self.hierarchy,
        )


INPUT_NODE = "input_node"
"""Name of the input node of a subgraph. Always created internally."""



import apache_beam as beam
from apache_beam.testing.test_pipeline import TestPipeline
from vulkan.core.policy import Policy
from vulkan.runners.beam.io import ReadLocalCSV
from vulkan.runners.beam.nodes import (
    BeamBranch,
    BeamInput,
    BeamTerminate,
    to_beam_nodes,
)
from vulkan.runners.beam.pipeline import build_pipeline
from vulkan.spec.dependency import INPUT_NODE, Dependency
from vulkan.spec.graph import sort_nodes
from vulkan.spec.nodes import BranchNode, NodeType, TerminateNode, TransformNode
from vulkan.spec.policy import PolicyDefinition


def test_pipeline():
    input_node = BeamInput(
        name=INPUT_NODE,
        schema={"score": int},
        data_path="vulkan/vulkan/runners/beam/test/input_data.csv",
    )

    def _branch(data):
        print(data)
        if data["score"] > 500:
            return "approved"
        return "denied"

    branch = BeamBranch(
        name="branch",
        func=_branch,
        choices=["approved", "denied"],
        dependencies={"data": Dependency(input_node.name)},
    )

    approved = BeamTerminate(
        name="approved",
        return_status="approved",
        dependencies={"condition": Dependency(branch.name, "approved")},
    )
    denied = BeamTerminate(
        name="denied",
        return_status="denied",
        dependencies={"condition": Dependency(branch.name, "denied")},
    )

    with TestPipeline() as pipeline:
        input_data = pipeline | "Read Input" >> ReadLocalCSV(input_node.data_path)
        collections = {INPUT_NODE: input_data}

        output, _ = build_pipeline(pipeline, collections, [branch, approved, denied])
        output | "Print" >> beam.Map(print)


def test_pipeline_from_policy():
    def _transform(data):
        data["p_score"] = data["score"] / 1000
        return data

    transform = TransformNode(
        name="transform",
        func=_transform,
        dependencies={"data": Dependency(INPUT_NODE)},
    )

    def _branch(data):
        if data["p_score"] > 0.5:
            return "approved"
        return "denied"

    branch = BranchNode(
        name="branch",
        func=_branch,
        choices=["approved", "denied"],
        dependencies={"data": Dependency(transform.name)},
    )

    approved = TerminateNode(
        name="approved",
        return_status="approved",
        dependencies={"condition": Dependency(branch.name, "approved")},
    )
    denied = TerminateNode(
        name="denied",
        return_status="denied",
        dependencies={"condition": Dependency(branch.name, "denied")},
    )

    definition = PolicyDefinition(
        nodes=[transform, branch, approved, denied],
        input_schema={"data": dict},
    )
    policy = Policy.from_definition(definition)
    nodes = policy.nodes
    edges = policy.edges

    input_node = [node for node in nodes if node.type == NodeType.INPUT][0]
    core_nodes = [node for node in nodes if node.type != NodeType.INPUT]

    beam_input = BeamInput.from_spec(
        input_node, data_path="vulkan/vulkan/runners/beam/test/input_data.csv"
    )
    beam_nodes = to_beam_nodes(core_nodes)
    sorted_nodes = sort_nodes(beam_nodes, edges)

    with TestPipeline() as pipeline:
        input_data = pipeline | "Read Input" >> ReadLocalCSV(beam_input.data_path)
        collections = {INPUT_NODE: input_data}

        output, _ = build_pipeline(pipeline, collections, sorted_nodes)
        output | "Print" >> beam.Map(print)



from copy import deepcopy

from vulkan.spec.dependency import Dependency
from vulkan.spec.nodes import (
    InputNode,
    Node,
    NodeType,
    TerminateNode,
    TransformNode,
)
from vulkan.spec.policy import PolicyDefinition


def resolve(
    nodes: list[Node],
) -> list[Node]:
    """Resolve a Policy Definition into a Policy.
    The resolution consists of a couple of things:
    1. Resolve any component node into their corresponding workflow;
    2. Join the outputs of each component into a single node;
    3. Remove the component nodes from the list of nodes;
    4. Update the dependencies for nodes that depend on components.
    """
    component_nodes = [node for node in nodes if node.type == NodeType.COMPONENT]
    # There is nothing to resolve if there are no nested flows (ie. components).
    if len(component_nodes) == 0:
        return nodes

    node_map = {node.id: node for node in nodes if node.type != NodeType.COMPONENT}

    # For each component, we resolve the component and then insert all
    # its nodes into the flattened node list.
    for node in component_nodes:
        if node.definition is None:
            msg = f"Component node {node.id} has no definition"
            raise ValueError(msg)

        hierarchy = (node.hierarchy or []) + [node.name]
        component = PolicyDefinition.from_dict(node.definition, hierarchy=hierarchy)
        component_nodes = {n.id: n for n in component.nodes}

        # Inject the component's dependencies into the input node
        # and replace the input node with an identity transform node.
        input_node = component.input_node.with_dependencies(node.dependencies)
        input_node = _identity_transform_from_input_node(input_node, node.parameters)
        component_nodes[input_node.id] = input_node

        # Replace the leaves (terminates) with identity transform nodes.
        leaves = {
            leaf.id: _identity_transform_from_terminate_node(leaf)
            for leaf in component.leaves.values()
        }
        component_nodes.update(leaves)

        # Join workflow outputs into an identity transform node
        output_joiner = _make_output_joiner(
            subpolicy_name=node.name,
            leaves=leaves,
            hierarchy=node.hierarchy,
        )
        component_nodes[output_joiner.id] = output_joiner

        # Add nodes in the component to the node map
        node_map.update(component_nodes)

    return list(node_map.values())


def _make_output_joiner(
    subpolicy_name: str,
    leaves: dict[str, Node],
    hierarchy: list[str] | None,
) -> TransformNode:
    description = f"Inserted automatically to join the outputs of {subpolicy_name}"
    dependencies = {
        name: Dependency(leaf.name, hierarchy=leaf.hierarchy)
        for name, leaf in leaves.items()
    }
    return TransformNode(
        name=subpolicy_name,
        func=_or_op_leaves,
        dependencies=dependencies,
        description=description,
        hierarchy=hierarchy,
    )


def _or_op_leaves(**kwargs):
    non_null_count = 0
    retval = None
    for value in kwargs.values():
        if value is not None:
            non_null_count += 1
            retval = value

    if non_null_count != 1:
        msg = f"Exactly one leaf should execute, got {non_null_count=}"
        raise ValueError(msg)
    return retval


def _identity_transform_from_input_node(
    input_node: InputNode,
    parameters: dict[str, str],
) -> TransformNode:
    return TransformNode(
        name=input_node.name,
        func=_identity_transform(input_node.schema),
        dependencies=input_node.dependencies,
        description=input_node.description,
        hierarchy=input_node.hierarchy,
        parameters=parameters,
    )


def _identity_transform_from_terminate_node(node: TerminateNode) -> TransformNode:
    parameters = node.output_data or {}
    parameters["return_status"] = node.return_status
    return TransformNode(
        name=node.name,
        func=_identity_fn,
        dependencies=node.dependencies,
        description=node.description,
        hierarchy=node.hierarchy,
        parameters=parameters,
    )


def _identity_fn(**kwargs):
    if len(kwargs) == 1:
        return list(kwargs.values())[0]
    return kwargs


def _identity_transform(input_schema: dict[str, type]):
    def _identity(**kwargs):
        if len(kwargs) == 1:
            return list(kwargs.values())[0]
        with_types = deepcopy(kwargs)
        for k, typ in input_schema.items():
            # If the key is in the input schema, we need to cast the value to the type.
            if k in with_types:
                with_types[k] = typ(with_types[k])
        return with_types

    return _identity



from typing import cast

from vulkan.spec.nodes.base import Node, NodeDefinition, NodeType
from vulkan.spec.nodes.metadata import ConnectionNodeMetadata


class ConnectionNode(Node):
    """A node that performs HTTP requests to external systems.

    Connection nodes are used to make HTTP requests to external APIs and services,
    allowing workflows to interact with web services and REST APIs.
    """

    def __init__(
        self,
        name: str,
        url: str,
        method: str = "GET",
        description: str | None = None,
        headers: dict | None = None,
        params: dict | None = None,
        body: dict | None = None,
        timeout: int | None = None,
        retry_max_retries: int = 1,
        response_type: str = "PLAIN_TEXT",
        dependencies: dict | None = None,
        hierarchy: list[str] | None = None,
    ):
        """Performs HTTP requests to external systems.

        Parameters
        ----------
        name : str
            The name of the node.
        url : str
            The URL to make the HTTP request to.
        method : str, optional
            The HTTP method to use (GET, POST, PUT, DELETE, etc.). Defaults to "GET".
        description : str, optional
            A description of the node.
        headers : dict, optional
            HTTP headers to include in the request.
        params : dict, optional
            Query parameters to include in the request.
        body : dict, optional
            Request body for POST/PUT requests.
        timeout : int, optional
            Request timeout in seconds.
        retry_max_retries : int, optional
            Maximum number of retry attempts. Defaults to 1.
        response_type : str, optional
            Expected response type. Defaults to "text/plain".
        dependencies : dict, optional
            The dependencies of the node.
        hierarchy : list[str], optional
            The hierarchy of the node.
        """
        super().__init__(
            name=name,
            description=description,
            typ=NodeType.CONNECTION,
            dependencies=dependencies,
            hierarchy=hierarchy,
        )
        self.url = url
        self.method = method
        self.headers = headers or {}
        self.params = params or {}
        self.body = body or {}
        self.timeout = timeout
        self.retry_max_retries = retry_max_retries
        self.response_type = response_type

    def node_definition(self) -> NodeDefinition:
        return NodeDefinition(
            name=self.name,
            description=self.description,
            node_type=self.type.value,
            dependencies=self.dependencies,
            metadata=ConnectionNodeMetadata(
                url=self.url,
                method=self.method,
                headers=self.headers,
                params=self.params,
                body=self.body,
                timeout=self.timeout,
                retry_max_retries=self.retry_max_retries,
                response_type=self.response_type,
            ),
            hierarchy=self.hierarchy,
        )

    @classmethod
    def from_dict(cls, spec: dict) -> "ConnectionNode":
        definition = NodeDefinition.from_dict(spec)
        if definition.metadata is None:
            raise ValueError(f"Metadata not set for node {definition.name}")

        metadata = cast(ConnectionNodeMetadata, definition.metadata)
        return cls(
            name=definition.name,
            description=definition.description,
            dependencies=definition.dependencies,
            url=metadata.url,
            method=metadata.method,
            headers=metadata.headers,
            params=metadata.params,
            body=metadata.body,
            timeout=metadata.timeout,
            retry_max_retries=metadata.retry_max_retries,
            response_type=metadata.response_type,
            hierarchy=definition.hierarchy,
        )



from vulkan.data_source import HTTPSource
from vulkan.node_config import RunTimeParam


class TestHTTPSource:
    def test_extract_runtime_params(self):
        config = HTTPSource(
            url="https://{{url}}",
            headers={
                "Authorization": {"param": "auth_token"},
                "TestHeader": {"param": "test_header", "value_type": "auto"},
                "accept": RunTimeParam(param="accept_type"),
                "Content-Type": RunTimeParam(param="content_type", value_type="str"),
            },
        )

        params = config.extract_runtime_params()

        assert set(params) == {
            "url",
            "auth_token",
            "test_header",
            "accept_type",
            "content_type",
        }

    def test_extract_runtime_params_with_escape_sequence(self):
        config = HTTPSource(
            url="http://url",
            method="POST",
            headers={
                "Content-Type": "application/json",
                "AccessToken": RunTimeParam(param="access_token", value_type="str"),
                "TokenId": RunTimeParam(param="token_id", value_type="str"),
            },
            body={
                "Datasets": "basic_data",
                "q": 'doc{{ "{" }}{{tax_id}}{{"}"}}',
                "Limit": 1,
            },
        )
        params = config.extract_runtime_params()
        assert set(params) == {"access_token", "token_id", "tax_id"}



from enum import Enum
from typing import Any, Literal

from pydantic import BaseModel


class BaseNodeMetadata(BaseModel):
    def to_dict(
        self, mode: str | Literal["json", "python"] = "python"
    ) -> dict[str, Any]:
        """Convert the metadata to a dictionary."""
        return self.model_dump(mode=mode)

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Any:
        """Load the metadata from a dictionary."""
        return cls.model_validate(data)


class InputNodeMetadata(BaseNodeMetadata):
    schema: dict[str, str]


class BranchNodeMetadata(BaseNodeMetadata):
    choices: list[str]
    source_code: str


class TransformNodeMetadata(BaseNodeMetadata):
    source_code: str
    parameters: dict[str, str] | None = None


class TerminateNodeMetadata(BaseNodeMetadata):
    return_status: str
    return_metadata: str | None = (
        None  # Generated from output_data for frontend compatibility
    )
    output_data: dict[str, str] | None = None  # Primary storage for SDK


class DataInputNodeMetadata(BaseNodeMetadata):
    data_source: str
    data_source_id: str | None = None
    parameters: dict[str, Any] | None = None


class PolicyNodeMetadata(BaseNodeMetadata):
    policy_id: str


class ConnectionNodeMetadata(BaseNodeMetadata):
    url: str
    method: str = "GET"
    headers: dict[str, Any] | None = None
    params: dict[str, Any] | None = None
    body: dict[str, Any] | None = None
    timeout: int | None = None
    retry_max_retries: int = 1
    response_type: str = "JSON"


class DecisionType(Enum):
    IF = "if"
    ELSE_IF = "else-if"
    ELSE = "else"


class DecisionCondition(BaseModel):
    decision_type: DecisionType
    condition: str | None = None
    output: str


class DecisionNodeMetadata(BaseNodeMetadata):
    conditions: list[DecisionCondition]


class ComponentNodeMetadata(BaseNodeMetadata):
    component_name: str
    definition: dict | None = None
    parameters: dict | None = None



from typing import Any

from vulkan.spec.dependency import Dependency
from vulkan.spec.nodes.base import Node, NodeDefinition, NodeType
from vulkan.spec.nodes.metadata import PolicyNodeMetadata


class PolicyDefinitionNode(Node):
    """A node that represents a policy definition.
    Policy nodes are used to "invoke" policies from within other policies.
    They're used to insert additional metadata for the policy so that it
    can be appropriately connectied to the rest of the workflow.
    """

    def __init__(
        self,
        name: str,
        policy_id: str,
        description: str | None = None,
        dependencies: dict[str, Dependency] | None = None,
    ):
        super().__init__(
            name=name,
            description=description,
            typ=NodeType.POLICY,
            dependencies=dependencies,
        )
        self.policy_id = policy_id

    def node_definition(self) -> NodeDefinition:
        metadata = PolicyNodeMetadata(policy_id=self.policy_id)
        return NodeDefinition(
            name=self.name,
            description=self.description,
            node_type=self.type.value,
            dependencies=self.dependencies,
            metadata=metadata,
        )

    @classmethod
    def from_dict(cls, spec: dict[str, Any]) -> "PolicyDefinitionNode":
        definition = NodeDefinition.from_dict(spec)
        if definition.node_type != NodeType.POLICY.value:
            raise ValueError(f"Expected NodeType.POLICY, got {definition.node_type}")
        if definition.metadata is None or definition.metadata.policy_id is None:
            raise ValueError("Missing policy metadata")

        return cls(
            name=definition.name,
            description=definition.description,
            dependencies=definition.dependencies,
            policy_id=definition.metadata.policy_id,
        )

    def to_dict(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "node_type": self.type.value,
            "description": self.description,
            "dependencies": self.dependencies,
            "metadata": {
                "policy_definition": self.policy_id,
            },
        }



from vulkan.spec.dependency import Dependency


def normalize_node_id(node_id: str) -> str:
    return node_id.replace(".", "_")


def normalize_dependencies(
    dependencies: dict[str, Dependency],
) -> dict[str, Dependency]:
    return {normalize_node_id(k): v for k, v in dependencies.items()}



"""
Authentication credential type constants for Data Sources.

This module defines the valid credential types that can be used
for authenticating with external data sources.
"""

from enum import Enum


class CredentialType(str, Enum):
    """Valid credential types for data source authentication."""

    CLIENT_ID = "CLIENT_ID"
    CLIENT_SECRET = "CLIENT_SECRET"
    USERNAME = "USERNAME"
    PASSWORD = "PASSWORD"


def validate_no_reserved_credentials_in_templates(
    env_var_names: set[str] | list[str],
    error_context: str = "templates",
) -> None:
    """
    Validates that reserved credential names are not used in templates

    Reserved credential names (CLIENT_ID, CLIENT_SECRET, USERNAME, PASSWORD)
    are restricted to the credentials endpoint and cannot be used in
    environment variables or templates

    Args:
        env_var_names: Set or list of environment variable names to check
        error_context: Context description for error message (default: "templates")

    Raises:
        ValueError: If any reserved credential names are found

    Example:
        validate_no_reserved_credentials_in_templates(["API_KEY", "CLIENT_ID"])
        ValueError: Reserved credential names cannot be used in templates: CLIENT_ID
    """
    if isinstance(env_var_names, list):
        env_var_names = set(env_var_names)

    conflicts = {name for name in env_var_names if name in CredentialType}
    if conflicts:
        raise ValueError(
            f"Reserved credential names cannot be used in {error_context}: {', '.join(sorted(conflicts))}. "
            f"These names are reserved for authentication only: {', '.join(CredentialType)}. "
            f"Please use the /credentials endpoint to set these values."
        )


def validate_credential_type(credential_type: str) -> None:
    """
    Validates that a credential type is valid

    Args:
        credential_type: Credential type to validate

    Raises:
        ValueError: If credential type is invalid

    Example:
        validate_credential_type("CLIENT_ID")
        validate_credential_type("INVALID")
        ValueError: Invalid credential_type: INVALID
    """
    if credential_type not in CredentialType:
        raise ValueError(
            f"Invalid credential_type: {credential_type}. "
            f"Must be one of: {', '.join(CredentialType)}"
        )



import json
from enum import Enum
from typing import Any, Callable, cast

from vulkan.node_config import (
    extract_env_vars_from_string,
    extract_runtime_params_from_string,
)
from vulkan.spec.dependency import Dependency
from vulkan.spec.nodes.base import Node, NodeDefinition, NodeType
from vulkan.spec.nodes.metadata import TerminateNodeMetadata


class TerminateNode(Node):
    """Marks the end of a workflow.

    Terminate nodes are used to mark the end of a workflow.
    They signal to the engine that the workflow has finished executing,
    and can be used to return a status code or a final decision, for example.

    Additionally, the user can specify a callback that will be executed
    when the node is run. This can be used to perform cleanup tasks, for example,
    or to communicate the final result of the workflow to an external system.

    All workflows must end with a terminate node, and all leaf nodes
    must be terminate nodes.
    This is currently not enforced, but it will be in the future.
    """

    def __init__(
        self,
        name: str,
        return_status: Enum | str,
        dependencies: dict[str, Dependency],
        description: str | None = None,
        callback: Callable | None = None,
        hierarchy: list[str] | None = None,
        output_data: dict[str, str] | None = None,
    ):
        """Marks the end of a workflow.

        Parameters
        ----------
        name : str
            The name of the node.
        return_status: Enum | str
            A "status" value that will be stored as the final status for the run.
        dependencies: dict, optional
            The dependencies of the node.
            See `Dependency` for more information.
        description: str, optional
            A description of the node.
        callback: Callable, optional
            A callback that will be executed when the node is run.
            In the current implementation, the callback function always
            receives an execution context as its first argument.
            TODO: improve documentation on callback function signature.
        output_data: dict[str, str], optional
            Output data dict with template support. Converted to return_metadata for storage.

        """
        self.return_status = (
            return_status.value if isinstance(return_status, Enum) else return_status
        )
        if dependencies is None:
            dependencies = {}

        super().__init__(
            name=name,
            description=description,
            typ=NodeType.TERMINATE,
            dependencies=dependencies,
            hierarchy=hierarchy,
        )
        self.callback = callback

        # Store output_data and convert to return_metadata for serialization
        self.output_data = output_data or {}
        self.return_metadata = (
            self._output_data_to_metadata(self.output_data)
            if self.output_data
            else None
        )

        # Validate templates
        if self.output_data:
            self._validate_output_data_templates(self.output_data)

    def _validate_json_metadata(self, json_metadata: str) -> None:
        """Validate JSON and templates."""
        try:
            parsed = json.loads(json_metadata)

            if not isinstance(parsed, dict):
                raise ValueError("JSON metadata must be a dictionary at the root level")

            def validate_value(value, path=""):
                if isinstance(value, str):
                    try:
                        extract_runtime_params_from_string(value)
                        extract_env_vars_from_string(value)
                    except ValueError as e:
                        raise ValueError(
                            f"Invalid template expression in JSON metadata at path '{path}': {value}. Error: {e}"
                        )

                elif isinstance(value, dict):
                    for k, v in value.items():
                        validate_value(v, f"{path}.{k}" if path else k)
                elif isinstance(value, list):
                    for i, item in enumerate(value):
                        validate_value(item, f"{path}[{i}]" if path else f"[{i}]")

            validate_value(parsed)

        except json.JSONDecodeError as e:
            raise ValueError(f"Invalid JSON in return_metadata: {e}")
        except Exception as e:
            if isinstance(e, ValueError):
                raise
            raise ValueError(f"Validation error in JSON metadata: {e}")

    def _validate_output_data_templates(self, output_data: dict[str, str]) -> None:
        """Validate templates in output_data."""
        for data_key, data_value in output_data.items():
            if isinstance(data_value, str):
                try:
                    extract_runtime_params_from_string(data_value)
                    extract_env_vars_from_string(data_value)
                except ValueError as e:
                    raise ValueError(
                        f"Invalid template expression in output_data at '{data_key}': {data_value}. Error: {e}"
                    )

    def _output_data_to_metadata(self, output_data: dict[str, str]) -> str:
        """Convert output_data to JSON."""
        try:
            return json.dumps(output_data, sort_keys=True)
        except (TypeError, ValueError) as e:
            raise ValueError(f"Failed to convert output_data to JSON: {e}")

    def node_definition(self) -> NodeDefinition:
        return NodeDefinition(
            name=self.name,
            description=self.description,
            node_type=self.type.value,
            dependencies=self.dependencies,
            metadata=TerminateNodeMetadata(
                return_status=self.return_status,
                return_metadata=self.return_metadata,
                output_data=self.output_data,
            ),
            hierarchy=self.hierarchy,
        )

    def with_callback(self, callback: Callable) -> "TerminateNode":
        self.callback = callback
        return self

    @classmethod
    def from_dict(cls, spec: dict[str, Any]) -> "TerminateNode":
        definition = NodeDefinition.from_dict(spec)
        if definition.metadata is None:
            raise ValueError(f"Metadata not set for TERMINATE node {definition.name}")

        metadata = cast(TerminateNodeMetadata, definition.metadata)

        # Priority: output_data > return_metadata
        if metadata.output_data:
            output_data = metadata.output_data
        elif metadata.return_metadata:
            output_data = cls._metadata_to_output_data_static(metadata.return_metadata)
        else:
            output_data = {}

        return cls(
            name=definition.name,
            description=definition.description,
            dependencies=definition.dependencies,
            return_status=metadata.return_status,
            output_data=output_data,
            hierarchy=definition.hierarchy,
        )

    @staticmethod
    def _metadata_to_output_data_static(return_metadata: str) -> dict[str, str]:
        """Convert JSON metadata to output_data dict for deserialization."""
        try:
            parsed = json.loads(return_metadata)
            if not isinstance(parsed, dict):
                return {"value": str(parsed)}
            return {k: str(v) for k, v in parsed.items()}
        except (json.JSONDecodeError, Exception):
            return {}



from enum import Enum

import pytest
from vulkan.spec.dependency import INPUT_NODE, Dependency
from vulkan.spec.nodes import BranchNode, InputNode, NodeType, TerminateNode
from vulkan.spec.nodes.transform import TransformNode


class ExampleReturnStatus(Enum):
    SUCCESS = "SUCCESS"


TEST_TABLE = {
    "Input Node - Simple": (
        InputNode,
        {
            "name": INPUT_NODE,
            "node_type": NodeType.INPUT.value,
            "metadata": {
                "schema": {
                    "cpf": "str",
                },
            },
        },
    ),
    "Input Node - Complete": (
        InputNode,
        {
            "name": INPUT_NODE,
            "node_type": NodeType.INPUT.value,
            "description": "Optional node Description",
            "metadata": {
                "schema": {
                    "cpf": "str",
                    "score": "int",
                    "valid": "bool",
                    "test": "float",
                },
            },
        },
    ),
    "Branch Node - Simple": (
        BranchNode,
        {
            "name": "branch",
            "node_type": NodeType.BRANCH.value,
            "dependencies": {},
            "metadata": {
                "choices": ["A", "B"],
                "source_code": """
return 10**2
            """,
            },
        },
    ),
    "Branch Node - With Import": (
        BranchNode,
        {
            "name": "branch",
            "node_type": NodeType.BRANCH.value,
            "description": "Optional node Description",
            "dependencies": {
                "node_a": Dependency("node_a"),
            },
            "metadata": {
                "choices": ["A", "B"],
                "source_code": """
import os
os.environ
            """,
            },
        },
    ),
    "Transform Node - Simple": (
        TransformNode,
        {
            "name": "transform",
            "node_type": NodeType.TRANSFORM.value,
            "dependencies": {},
            "metadata": {
                "source_code": """
return 10**2
            """,
            },
        },
    ),
    "Transform Node - With Import": (
        TransformNode,
        {
            "name": "transform",
            "node_type": NodeType.TRANSFORM.value,
            "description": "Optional node Description",
            "dependencies": {
                "node_a": Dependency("node_a"),
            },
            "metadata": {
                "source_code": """
import os
os.environ
            """,
            },
        },
    ),
    "Terminate Node - Simple": (
        TerminateNode,
        {
            "name": "terminate_a",
            "node_type": NodeType.TERMINATE.value,
            "metadata": {
                "return_status": "SUCCESS",
            },
            "dependencies": {
                "node_a": Dependency("node_a"),
            },
        },
    ),
    "Terminate Node - With Enum": (
        TerminateNode,
        {
            "name": "terminate_b",
            "node_type": NodeType.TERMINATE.value,
            "description": "Optional node Description",
            "metadata": {
                "return_status": ExampleReturnStatus.SUCCESS,
            },
            "dependencies": {
                "node_a": Dependency("node_a"),
            },
        },
    ),
    "Terminate Node - With Parameters": (
        TerminateNode,
        {
            "name": "terminate_with_params",
            "node_type": NodeType.TERMINATE.value,
            "description": "Terminate node with parameters",
            "metadata": {
                "return_status": "COMPLETED",
                "output_data": {
                    "message": "Task completed successfully",
                    "timestamp": "{{input.timestamp}}",
                    "user_id": "{{input.user_id}}",
                },
            },
            "dependencies": {
                "input": Dependency("input_node"),
            },
        },
    ),
    "Terminate Node - Mixed Parameters and Metadata": (
        TerminateNode,
        {
            "name": "terminate_mixed",
            "node_type": NodeType.TERMINATE.value,
            "description": "Terminate node with both parameters and metadata",
            "metadata": {
                "return_status": "SUCCESS",
                "return_metadata": '{"legacy_field": "{{input.data}}"}',
                "output_data": {
                    "new_field": "{{input.new_data}}",
                    "static_value": "constant",
                },
            },
            "dependencies": {
                "input": Dependency("input_node"),
            },
        },
    ),
    # TODO: Data input tests
}


@pytest.mark.parametrize(
    ["node_cls", "spec"],
    [(cls, spec) for cls, spec in TEST_TABLE.values()],
    ids=list(TEST_TABLE.keys()),
)
def test_node_from_spec(node_cls, spec):
    node = node_cls.from_dict(spec)
    assert node.name == spec["name"]
    assert node.type.value == spec["node_type"]
    assert node.description == spec.get("description", None)


def test_terminate_node_parameters():
    """Test parameters handling."""
    # Create with parameters
    node = TerminateNode(
        name="test_terminate",
        return_status="SUCCESS",
        dependencies={"input": Dependency("input_node")},
        output_data={
            "message": "Test completed",
            "template_value": "{{input.data}}",
            "static_number": "42",
        },
    )

    # Verify storage
    assert node.output_data == {
        "message": "Test completed",
        "template_value": "{{input.data}}",
        "static_number": "42",
    }

    # Check serialization
    node_def = node.node_definition()
    assert node_def.metadata.output_data == node.output_data

    # Round-trip test
    spec_dict = node_def.to_dict()
    reconstructed_node = TerminateNode.from_dict(spec_dict)

    # Verify round-trip
    assert reconstructed_node.output_data == node.output_data
    assert reconstructed_node.name == node.name
    assert reconstructed_node.return_status == node.return_status


def test_terminate_node_parameter_to_metadata_conversion():
    """Test parameter to return_metadata conversion."""
    # Parameters are converted to return_metadata for serialization
    node = TerminateNode(
        name="test_terminate",
        return_status="SUCCESS",
        dependencies={"input": Dependency("input_node")},
        output_data={"key": "{{input.data}}", "value": "test"},
    )

    # Parameters stored as-is, return_metadata generated as JSON
    assert node.output_data == {"key": "{{input.data}}", "value": "test"}
    assert node.return_metadata == '{"key": "{{input.data}}", "value": "test"}'


def test_terminate_node_dependencies_validation():
    """Test dependencies validation."""
    node = TerminateNode(
        name="test_terminate",
        return_status="SUCCESS",
        dependencies={"input": Dependency("input_node")},
    )
    assert node.name == "test_terminate"
    assert len(node.dependencies) == 1


def test_terminate_node_parameters_validation():
    """Test parameters validation."""
    # Valid parameters should work
    node = TerminateNode(
        name="test_terminate",
        return_status="SUCCESS",
        dependencies={"input": Dependency("input_node")},
        output_data={"key": "value", "template": "{{input.data}}"},
    )
    assert node.output_data == {"key": "value", "template": "{{input.data}}"}
    assert node.return_metadata == '{"key": "value", "template": "{{input.data}}"}'


def test_terminate_node_template_validation():
    """Test template validation."""
    # Invalid template -> error
    with pytest.raises(ValueError, match="Invalid template expression"):
        TerminateNode(
            name="test_terminate",
            return_status="SUCCESS",
            dependencies={"input": Dependency("input_node")},
            output_data={
                "invalid_template": "{{unclosed.template"
            },  # Missing closing braces
        )

    # Valid templates -> ok
    node = TerminateNode(
        name="test_terminate",
        return_status="SUCCESS",
        dependencies={"input": Dependency("input_node")},
        output_data={"valid": "{{input.data}}", "also_valid": "{{input.nested.value}}"},
    )
    assert node.output_data is not None
    assert node.return_metadata is not None


def test_terminate_node_enum_return_status():
    """Test Enum return_status handling."""
    # Enum input
    node = TerminateNode(
        name="test_terminate",
        return_status=ExampleReturnStatus.SUCCESS,
        dependencies={"input": Dependency("input_node")},
    )

    # Extract string value
    assert node.return_status == "SUCCESS"
    assert isinstance(node.return_status, str)

    # Serialization preserves string
    node_def = node.node_definition()
    assert node_def.metadata.return_status == "SUCCESS"


def test_terminate_node_serialization_edge_cases():
    """Test serialization edge cases."""
    # Minimal node
    node_minimal = TerminateNode(
        name="minimal_terminate",
        return_status="DONE",
        dependencies={"input": Dependency("input_node")},
    )

    # Check minimal serialization
    spec_dict = node_minimal.node_definition().to_dict()
    assert spec_dict["name"] == "minimal_terminate"
    assert spec_dict["metadata"]["return_status"] == "DONE"
    assert spec_dict["metadata"].get("return_metadata") is None
    assert spec_dict["metadata"].get("output_data") == {}

    # Check minimal deserialization
    reconstructed = TerminateNode.from_dict(spec_dict)
    assert reconstructed.name == "minimal_terminate"
    assert reconstructed.return_status == "DONE"
    assert reconstructed.return_metadata is None  # No parameters = no return_metadata
    assert reconstructed.output_data == {}

    # Full node
    node_full = TerminateNode(
        name="full_terminate",
        return_status="COMPLETE",
        dependencies={"input": Dependency("input_node")},
        description="Full featured terminate node",
        output_data={"extra": "data", "message": "All done"},
    )

    # Full round-trip
    spec_dict_full = node_full.node_definition().to_dict()
    reconstructed_full = TerminateNode.from_dict(spec_dict_full)

    assert reconstructed_full.name == node_full.name
    assert reconstructed_full.return_status == node_full.return_status
    assert reconstructed_full.description == node_full.description
    assert reconstructed_full.return_metadata == node_full.return_metadata
    assert reconstructed_full.output_data == node_full.output_data


def test_terminate_node_from_dict_error_cases():
    """Test from_dict error cases."""
    # Missing metadata -> error
    invalid_spec_no_metadata = {
        "name": "test_terminate",
        "node_type": NodeType.TERMINATE.value,
        "dependencies": {"input": Dependency("input_node")},
        # Missing metadata key
    }

    with pytest.raises(ValueError, match="Metadata not set for TERMINATE node"):
        TerminateNode.from_dict(invalid_spec_no_metadata)

    # None metadata -> error
    invalid_spec_none_metadata = {
        "name": "test_terminate",
        "node_type": NodeType.TERMINATE.value,
        "dependencies": {"input": Dependency("input_node")},
        "metadata": None,
    }

    with pytest.raises(ValueError, match="Metadata not set for TERMINATE node"):
        TerminateNode.from_dict(invalid_spec_none_metadata)


def test_terminate_node_callback_handling():
    """Test callback handling."""
    # with_callback method
    node = TerminateNode(
        name="test_terminate",
        return_status="SUCCESS",
        dependencies={"input": Dependency("input_node")},
    )

    def test_callback(context):
        return "callback executed"

    # Returns self for chaining
    result = node.with_callback(test_callback)
    assert result is node

    # Callback is set
    assert node.callback == test_callback

    # Constructor callback
    node_with_callback = TerminateNode(
        name="test_terminate_cb",
        return_status="SUCCESS",
        dependencies={"input": Dependency("input_node")},
        callback=test_callback,
    )

    assert node_with_callback.callback == test_callback


def test_terminate_node_simple_parameters():
    """Test simple parameter handling."""
    # Simple parameters should work
    node = TerminateNode(
        name="simple_terminate",
        return_status="DONE",
        dependencies={"input": Dependency("input_node")},
        output_data={
            "status": "{{decision.final_status}}",
            "score": "{{risk.calculated_score}}",
            "static_info": "This is static",
        },
    )

    assert node.output_data == {
        "status": "{{decision.final_status}}",
        "score": "{{risk.calculated_score}}",
        "static_info": "This is static",
    }

    # Round-trip should preserve parameters
    spec_dict = node.node_definition().to_dict()
    reconstructed = TerminateNode.from_dict(spec_dict)
    assert reconstructed.output_data == node.output_data


def test_terminate_node_parameters_template_validation():
    """Test parameters template validation."""
    # Invalid parameters template -> error
    with pytest.raises(ValueError, match="Invalid template expression"):
        TerminateNode(
            name="test_terminate",
            return_status="SUCCESS",
            dependencies={"input": Dependency("input_node")},
            output_data={
                "invalid": "{{bad..syntax}}"
            },  # Invalid template in parameters
        )

    # Valid parameters templates -> ok
    node = TerminateNode(
        name="test_terminate",
        return_status="SUCCESS",
        dependencies={"input": Dependency("input_node")},
        output_data={
            "valid_simple": "{{input.data}}",
            "valid_nested": "{{input.nested.field}}",
            "static_value": "no template here",
        },
    )
    assert len(node.output_data) == 3


def test_terminate_node_return_status_validation():
    """Test return_status edge cases."""
    # None return_status allowed
    node_none_status = TerminateNode(
        name="test_terminate",
        return_status=None,
        dependencies={"input": Dependency("input_node")},
    )
    assert node_none_status.return_status is None

    # Empty string allowed
    node_empty_status = TerminateNode(
        name="test_terminate",
        return_status="",
        dependencies={"input": Dependency("input_node")},
    )
    assert node_empty_status.return_status == ""

    # Valid string works
    node_valid = TerminateNode(
        name="test_terminate",
        return_status="COMPLETED",
        dependencies={"input": Dependency("input_node")},
    )
    assert node_valid.return_status == "COMPLETED"


def test_terminate_node_empty_collections():
    """Test empty values handling."""
    # Empty parameters dict works
    node_empty_params = TerminateNode(
        name="test_terminate",
        return_status="SUCCESS",
        dependencies={"input": Dependency("input_node")},
        output_data={},
    )
    assert node_empty_params.output_data == {}
    assert node_empty_params.return_metadata is None  # No metadata for empty parameters

    # None parameters works (defaults to empty)
    node_no_params = TerminateNode(
        name="test_terminate",
        return_status="SUCCESS",
        dependencies={"input": Dependency("input_node")},
    )
    assert node_no_params.output_data == {}
    assert node_no_params.return_metadata is None


def test_terminate_node_parameters_mixed_types():
    """Test parameters type handling."""
    # All values should be strings
    node = TerminateNode(
        name="test_terminate",
        return_status="SUCCESS",
        dependencies={"input": Dependency("input_node")},
        output_data={
            "string_value": "hello",
            "number_as_string": "42",
            "boolean_as_string": "true",
            "template": "{{input.value}}",
            "json_like": '{"nested": "value"}',
        },
    )

    # Verify all strings
    assert all(isinstance(v, str) for v in node.output_data.values())
    assert node.output_data["number_as_string"] == "42"
    assert node.output_data["boolean_as_string"] == "true"
    assert node.output_data["json_like"] == '{"nested": "value"}'

    # Round-trip preserves types
    spec_dict = node.node_definition().to_dict()
    reconstructed = TerminateNode.from_dict(spec_dict)
    assert reconstructed.output_data == node.output_data


def test_terminate_node_parameter_to_metadata_serialization():
    """Test parameters to return_metadata conversion for serialization."""

    # SDK path: parameters -> return_metadata
    sdk_node = TerminateNode(
        name="sdk_node",
        return_status="SUCCESS",
        dependencies={"input": Dependency("input_node")},
        output_data={
            "message": "User approved",
            "score": "85",
            "template": "{{input.data}}",
        },
    )

    # Auto-generate JSON from parameters
    assert (
        sdk_node.return_metadata
        == '{"message": "User approved", "score": "85", "template": "{{input.data}}"}'
    )
    assert sdk_node.output_data == {
        "message": "User approved",
        "score": "85",
        "template": "{{input.data}}",
    }

    # Round-trip through serialization
    spec_dict = sdk_node.node_definition().to_dict()
    reconstructed = TerminateNode.from_dict(spec_dict)
    assert reconstructed.output_data == sdk_node.output_data


def test_terminate_node_backward_compatibility_deserialization():
    """Test deserializing old return_metadata format."""
    # Simulate old serialized format with return_metadata
    old_spec = {
        "name": "legacy_node",
        "node_type": "TERMINATE",
        "dependencies": {
            "input": {
                "node": "input_node",
                "output": None,
                "key": None,
                "hierarchy": None,
            }
        },
        "metadata": {
            "return_status": "SUCCESS",
            "return_metadata": '{"user": "John", "score": "95"}',
        },
    }

    # Should be able to deserialize and convert to parameters
    node = TerminateNode.from_dict(old_spec)
    assert node.output_data == {"user": "John", "score": "95"}
    # JSON is sorted by keys
    assert node.return_metadata == '{"score": "95", "user": "John"}'


def test_terminate_node_parameters_only():
    """Test that only parameters are accepted in constructor."""

    # Only parameters are used now
    node = TerminateNode(
        name="params_node",
        return_status="SUCCESS",
        dependencies={"input": Dependency("input_node")},
        output_data={"source": "sdk", "value": "test"},
    )

    # Verify parameters are stored and converted to return_metadata
    assert node.output_data == {"source": "sdk", "value": "test"}
    assert node.return_metadata == '{"source": "sdk", "value": "test"}'


def test_terminate_node_parameter_type_handling():
    """Test parameter type handling."""

    # All parameter values should be strings
    node = TerminateNode(
        name="typed_node",
        return_status="SUCCESS",
        dependencies={"input": Dependency("input_node")},
        output_data={"string": "text", "template": "{{input.value}}"},
    )

    # Verify all stored as strings
    assert node.output_data == {"string": "text", "template": "{{input.value}}"}
    assert node.return_metadata == '{"string": "text", "template": "{{input.value}}"}'

    # Test deserialization from metadata with mixed types
    old_spec = {
        "name": "mixed_node",
        "node_type": "TERMINATE",
        "dependencies": {
            "input": {
                "node": "input_node",
                "output": None,
                "key": None,
                "hierarchy": None,
            }
        },
        "metadata": {
            "return_status": "SUCCESS",
            "return_metadata": '{"number": 42, "boolean": true, "string": "text"}',
        },
    }

    deserialized = TerminateNode.from_dict(old_spec)
    # Should convert all to strings (Python uses "True" for booleans)
    assert deserialized.output_data == {
        "number": "42",
        "boolean": "True",
        "string": "text",
    }



import os

import yaml
from dagster import Definitions, EnvVar, IOManagerDefinition

from vulkan.runners.dagster.io_manager import (
    DB_CONFIG_KEY,
    DBConfig,
    postgresql_io_manager,
)
from vulkan.runners.dagster.policy import DagsterFlow
from vulkan.runners.dagster.resources import (
    AppClientResource,
    VulkanPolicyConfig,
    VulkanRunConfig,
)
from vulkan.runners.shared.constants import (
    APP_CLIENT_KEY,
    POLICY_CONFIG_KEY,
    RUN_CONFIG_KEY,
)
from vulkan.spec.environment.loaders import load_and_resolve_policy


def make_workspace_definition(spec_file_path: str) -> Definitions:
    run_config = VulkanRunConfig.configure_at_launch()
    resources = {
        # Vulkan Configurable Resources
        RUN_CONFIG_KEY: run_config,
        POLICY_CONFIG_KEY: VulkanPolicyConfig.configure_at_launch(),
        # Run DB
        DB_CONFIG_KEY: DBConfig(
            host=EnvVar("VULKAN_DB_HOST"),
            port=EnvVar.int("VULKAN_DB_PORT"),
            user=EnvVar("VULKAN_DB_USER"),
            password=EnvVar("VULKAN_DB_PASSWORD"),
            database=EnvVar("VULKAN_DB_DATABASE"),
            object_table=EnvVar("VULKAN_DB_OBJECT_TABLE"),
        ),
        # Unified App Client
        APP_CLIENT_KEY: AppClientResource(run_config=run_config),
        # IO Managers
        "io_manager": IOManagerDefinition(
            resource_fn=postgresql_io_manager,
            required_resource_keys={RUN_CONFIG_KEY, DB_CONFIG_KEY},
        ),
    }

    # Up to this point, everything should be defined in terms of core elements.
    # Nodes and components should be configured, resolved, checked in core.
    resolved_policy = load_and_resolve_policy(spec_file_path)
    # From here, each implementation should handle transforming core to its own
    # needs, ie. Core -> Dagster
    # -> Transform nodes in dagster nodes
    dagster_flow = DagsterFlow(nodes=resolved_policy.nodes)
    # By definition, Vulkan dagster worskpaces have a single job.
    jobs = [dagster_flow.to_job(resources)]
    definition = Definitions(
        assets=[],
        jobs=jobs,
        resources={},
    )
    return definition


class DagsterWorkspaceManager:
    def __init__(
        self,
        base_dir: str,
        workspace_path: str,
    ) -> None:
        self.base_dir = base_dir
        self.workspace_path = workspace_path

    def add_workspace_config(self, workspace_id: str):
        workspace_file = os.path.join(self.base_dir, "workspace.yaml")
        with open(workspace_file, "r") as fp:
            workspace_config = yaml.safe_load(fp)

        for cfg in workspace_config["load_from"]:
            if cfg.get("python_file", {}).get("location_name") == workspace_id:
                return

        # If we get here, the workspace_id is not in the config file
        new_config = {
            "python_file": {
                "relative_path": f"{self.workspace_path}/__init__.py",
                "working_directory": self.workspace_path,
                "executable_path": f"{self.workspace_path}/.venv/bin/python",
                "location_name": workspace_id,
            }
        }
        workspace_config["load_from"].append(new_config)
        with open(workspace_file, "w") as fp:
            yaml.dump(workspace_config, fp)

    def remove_workspace_config(self, workspace_id: str):
        path = os.path.join(self.base_dir, "workspace.yaml")

        with open(path, "r") as fn:
            workspace_config = yaml.safe_load(fn)

        load_from = []
        for location in workspace_config["load_from"]:
            if location.get("python_file", {}).get("location_name") != workspace_id:
                load_from.append(location)

        with open(path, "w") as fn:
            yaml.dump(dict(load_from=load_from), fn)

    def create_init_file(self) -> str:
        init_path = os.path.join(self.workspace_path, "__init__.py")

        with open(init_path, "w") as fp:
            init_contents = DAGSTER_ENTRYPOINT.format(
                spec_file_path=os.path.join(self.workspace_path, "policy.json"),
            )
            fp.write(init_contents)

        return init_path

    def delete_resources(self, workspace_id: str):
        self.remove_workspace_config(workspace_id)


DAGSTER_ENTRYPOINT = """
from vulkan.runners.dagster.workspace import make_workspace_definition

definitions = make_workspace_definition("{spec_file_path}")
"""



import logging
import os

from hatchet_sdk import Hatchet
from hatchet_sdk.config import ClientConfig
from pydantic import BaseModel

from vulkan.runners.hatchet.policy import HatchetFlow
from vulkan.spec.environment.loaders import load_and_resolve_policy

logger = logging.getLogger(__name__)


class HatchetClientConfig(BaseModel):
    token: str
    host_port: str | None = None
    server_url: str | None = None
    namespace: str | None = None

    @classmethod
    def from_env(cls) -> "HatchetClientConfig":
        """Create configuration from environment variables."""
        token = os.getenv("HATCHET_CLIENT_TOKEN")
        if not token:
            raise ValueError("HATCHET_CLIENT_TOKEN is not set")

        host_port = os.getenv("HATCHET_CLIENT_HOST_PORT", "hatchet-engine:7070")
        server_url = os.getenv("HATCHET_SERVER_URL", None)
        namespace = os.getenv("HATCHET_NAMESPACE", "")

        return cls(
            host_port=host_port,
            server_url=server_url,
            token=token,
            namespace=namespace,
        )


class HatchetWorkspaceManager:
    """Manages Hatchet workspace, workers, and workflow execution."""

    def __init__(
        self,
        hatchet_client_config: HatchetClientConfig,
        spec_file_path: str,
        workflow_id: str,
    ):
        if not os.path.exists(spec_file_path):
            raise ValueError(f"Spec file does not exist: {spec_file_path}")

        self.workflow_id = workflow_id
        self.config = hatchet_client_config
        server_url = (
            self.config.server_url
            if self.config.server_url
            else ClientConfig().server_url
        )
        self.client = Hatchet(
            config=ClientConfig(
                server_url=server_url,
                host_port=self.config.host_port,
                token=self.config.token,
                namespace=self.config.namespace,
            ),
        )

        policy = load_and_resolve_policy(spec_file_path)
        hatchet_flow = HatchetFlow(policy.nodes, policy_name=self.workflow_id)
        self._hatchet_workflow = hatchet_flow.create_workflow()
        # For now, we're creating one worker per workflow/workspace.
        self._worker = self.client.worker(
            f"vulkan-worker-{self.workflow_id}",
            workflows=[self._hatchet_workflow],
        )

        logger.info(
            f"Initialized Hatchet workspace for namespace: {self.config.namespace}"
        )

    def start_worker(self) -> None:
        """Start the Hatchet worker."""
        try:
            logger.info("Starting Hatchet worker...")
            self._worker.start()
        except Exception as e:
            logger.error(f"Failed to start worker: {e}")
            raise

    def stop_worker(self) -> None:
        """Stop the Hatchet worker."""
        try:
            self._worker.stop()
            logger.info("Stopped Hatchet worker")
        except Exception as e:
            logger.error(f"Failed to stop worker: {e}")

    def cleanup(self) -> None:
        """Clean up resources."""
        self.stop_worker()
        self.client = None
        self._worker = None


HATCHET_ENTRYPOINT = """
from vulkan.runners.hatchet.workspace import HatchetWorkspaceManager, HatchetClientConfig

if __name__ == "__main__":
    hatchet_client_config = HatchetClientConfig.from_env()
    manager = HatchetWorkspaceManager(
        hatchet_client_config=hatchet_client_config,
        spec_file_path="{spec_file_path}",
        workflow_id="{workflow_id}",
    )

    manager.start_worker()
"""



import json
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Callable

import httpx
from dagster import In, OpDefinition, OpExecutionContext, Out, Output

from vulkan.connections import (
    HTTPConfig,
    RetryPolicy,
    format_response_data,
)
from vulkan.core.context import VulkanExecutionContext
from vulkan.core.run import RunStatus
from vulkan.exceptions import UserCodeException
from vulkan.http_client import HTTPClient
from vulkan.node_config import resolve_template, resolve_value
from vulkan.runners.dagster.names import normalize_dependencies, normalize_node_id
from vulkan.runners.dagster.resources import AppClientResource
from vulkan.runners.shared.constants import (
    APP_CLIENT_KEY,
    POLICY_CONFIG_KEY,
    RUN_CONFIG_KEY,
)
from vulkan.runners.shared.decision_fn import evaluate_condition
from vulkan.runners.shared.run_config import VulkanPolicyConfig, VulkanRunConfig
from vulkan.spec.dependency import Dependency
from vulkan.spec.nodes import (
    BranchNode,
    ConnectionNode,
    DataInputNode,
    DecisionNode,
    InputNode,
    Node,
    TerminateNode,
    TransformNode,
)
from vulkan.spec.nodes.metadata import DecisionCondition, DecisionType
from vulkan.spec.policy import PolicyDefinitionNode


class DagsterNode(ABC):
    @abstractmethod
    def op(self) -> OpDefinition:
        """Construct the Dagster op for this node."""


class DagsterDataInput(DataInputNode, DagsterNode):
    def __init__(
        self,
        name: str,
        data_source: str,
        description: str | None = None,
        parameters: dict[str, str] | None = None,
        dependencies: dict | None = None,
    ):
        super().__init__(
            name=name,
            data_source=data_source,
            description=description,
            parameters=parameters,
            dependencies=dependencies,
        )

    def op(self) -> OpDefinition:
        return OpDefinition(
            compute_fn=self.run,
            name=self.name,
            ins={k: In() for k in self.dependencies.keys()},
            outs={
                "result": Out(),
            },
            required_resource_keys={APP_CLIENT_KEY, POLICY_CONFIG_KEY, RUN_CONFIG_KEY},
        )

    def run(self, context: OpExecutionContext, inputs: dict[str, Any]):
        app_client_resource: AppClientResource = getattr(
            context.resources, APP_CLIENT_KEY
        )
        client = app_client_resource.get_client()
        inputs = _resolved_inputs(inputs, self.dependencies)

        try:
            configured_params = self._get_configured_params(inputs)
            context.log.info(
                f"Fetching data from data source {self.data_source} with "
                f"parameters: {configured_params}"
            )

            response = client.fetch_data(
                data_source=self.data_source,
                configured_params=configured_params,
            )

            response.raise_for_status()

            data = response.json()
            context.log.info(f"Data: {data}")

            yield Output(data["value"])

        except ValueError as e:
            context.log.error(f"Parameter resolution error: {str(e)}")
            raise e

        except httpx.HTTPStatusError as e:
            error_detail = "Unknown error"
            if hasattr(e, "response") and e.response is not None:
                response = e.response
                try:
                    error_detail = response.json().get("detail", str(e))
                except Exception:
                    error_detail = (
                        response.text[:200] if hasattr(response, "text") else str(e)
                    )
                context.log.error(
                    f"Failed request with status {response.status_code}: {error_detail}"
                )
            else:
                context.log.error(f"HTTP error: {str(e)}")

            raise e

        except httpx.HTTPError as e:
            context.log.error(f"Failed to retrieve data: {str(e)}")
            raise e

        except Exception as e:
            context.log.error(f"Unexpected error processing data: {str(e)}")
            raise e

    def _get_configured_params(self, inputs):
        if self.parameters is None:
            return {}

        configured_params = {}
        for k, v in self.parameters.items():
            try:
                configured_params[k] = resolve_value(v, inputs, env_variables={})
            except Exception:
                raise ValueError(f"Invalid parameter configuration: {v}")
        return configured_params

    @classmethod
    def from_spec(cls, node: DataInputNode):
        return cls(
            name=normalize_node_id(node.id),
            data_source=node.data_source,
            description=node.description,
            parameters=node.parameters,
            dependencies=node.dependencies,
        )


class DagsterPolicy(PolicyDefinitionNode, DagsterNode):
    def __init__(
        self,
        name: str,
        policy_id: str,
        dependencies: dict | None = None,
    ):
        super().__init__(
            name=name,
            policy_id=policy_id,
            dependencies=dependencies,
        )

    def op(self) -> OpDefinition:
        return OpDefinition(
            compute_fn=self.run,
            name=self.name,
            ins={k: In() for k in self.dependencies.keys()},
            outs={
                "result": Out(),
            },
            required_resource_keys={APP_CLIENT_KEY, POLICY_CONFIG_KEY, RUN_CONFIG_KEY},
        )

    def run(self, context: OpExecutionContext, inputs: dict[str, Any]):
        app_client_resource: AppClientResource = getattr(
            context.resources, APP_CLIENT_KEY
        )
        client = app_client_resource.get_client()
        inputs = _resolved_inputs(inputs, self.dependencies)

        # TODO: handle schema same way as data input nodes
        input_data = inputs.get("body", None)
        body = {"input_data": input_data}

        try:
            result = client.run_version_sync(
                policy_version_id=self.policy_id,
                data=body,
            )
            yield Output(result)
        except ValueError as e:
            context.log.error(f"Failed op {self.name}: {e}")
            raise e

    @classmethod
    def from_spec(cls, node: PolicyDefinitionNode):
        return cls(
            name=normalize_node_id(node.id),
            policy_id=node.policy_id,
            dependencies=node.dependencies,
        )


class DagsterTransformNodeMixin(DagsterNode):
    def op(self) -> OpDefinition:
        def fn(context, inputs):
            inputs = _resolved_inputs(inputs, self.dependencies)
            configured_params = self._get_configured_params(inputs)
            try:
                fn_params = {**inputs, **configured_params}
                result = self._func(context, **fn_params)
                yield Output(result, output_name="result")
            except Exception as e:
                raise UserCodeException(self.name) from e

        node_op = OpDefinition(
            compute_fn=fn,
            name=self.name,
            ins={k: In() for k in self.dependencies.keys()},
            outs={
                "result": Out(),
            },
            # We expose the configuration in transform nodes
            # to allow the callback function in terminate nodes to
            # access it. In the future, we may separate terminate nodes.
            required_resource_keys={APP_CLIENT_KEY, RUN_CONFIG_KEY, POLICY_CONFIG_KEY},
        )

        return node_op

    def _get_configured_params(self, inputs):
        params = getattr(self, "parameters", None)
        if params is None:
            return {}

        configured_params = {}
        for k, v in params.items():
            try:
                configured_params[k] = resolve_value(v, inputs, env_variables={})
            except Exception:
                raise ValueError(f"Invalid parameter configuration: {v}")
        return configured_params


def _with_vulkan_context(func: callable) -> callable:
    def fn(context: OpExecutionContext, **kwargs):
        if func.__code__.co_varnames[0] == "context":
            env = getattr(context.resources, POLICY_CONFIG_KEY)
            ctx = VulkanExecutionContext(logger=context.log, env=env.variables)
            return func(ctx, **kwargs)
        return func(**kwargs)

    return fn


def _resolved_inputs(
    inputs: dict[str, Any], dependencies: dict[str, Any]
) -> dict[str, Any]:
    resolved_inputs = {}
    for k, v in inputs.items():
        dep = dependencies.get(k)
        if dep is not None and dep.key is not None:
            resolved_inputs[k] = v[dep.key]
        else:
            resolved_inputs[k] = v
    return resolved_inputs


class DagsterTransform(TransformNode, DagsterTransformNodeMixin):
    def __init__(
        self,
        name: str,
        description: str,
        func: Callable,
        dependencies: dict[str, Any],
        parameters: dict[str, str] | None = None,
    ):
        super().__init__(
            name=name,
            description=description,
            func=func,
            dependencies=dependencies,
            parameters=parameters,
        )
        self._func = _with_vulkan_context(self.func)

    @classmethod
    def from_spec(cls, node: TransformNode):
        # This is only necessary here, as TransformNodes are the only
        # type that can depend on internal nodes.
        deps = normalize_dependencies(node.dependencies)
        return cls(
            name=normalize_node_id(node.id),
            description=node.description,
            func=node.func,
            dependencies=deps,
            parameters=node.parameters,
        )


class DagsterTerminate(TerminateNode, DagsterTransformNodeMixin):
    def __init__(
        self,
        name: str,
        description: str,
        return_status: str,
        dependencies: dict[str, Dependency],
        output_data: dict[str, str] | None = None,
        callback: Callable | None = None,
    ):
        super().__init__(
            name=name,
            description=description,
            return_status=return_status,
            output_data=output_data,
            dependencies=dependencies,
            callback=callback,
        )
        self._func = self._fn

    def _fn(self, context: OpExecutionContext, **kwargs):
        status = self.return_status
        result = status.value if isinstance(status, Enum) else status
        vulkan_run_config: VulkanRunConfig = getattr(context.resources, RUN_CONFIG_KEY)
        context.log.debug(f"Terminating with status {status}")

        metadata = None
        if self.return_metadata is not None:
            try:
                template_metadata = json.loads(self.return_metadata)
                metadata = self._resolve_json_metadata(
                    template_metadata, kwargs, context
                )
            except json.JSONDecodeError as e:
                context.log.error(f"Failed to parse JSON metadata: {e}")
                raise ValueError(f"Invalid JSON in return_metadata: {e}")
            except Exception as e:
                context.log.error(f"Failed to resolve JSON metadata: {e}")
                raise ValueError(f"Failed to resolve JSON metadata: {e}")

        terminated = self._terminate(context, result, metadata)
        if not terminated:
            raise ValueError("Failed to terminate run")

        if self.callback is not None:
            reported = self.callback(
                context=context,
                run_id=vulkan_run_config.run_id,
                return_status=status,
                **kwargs,
            )
            if not reported:
                raise ValueError("Callback function failed")

        return result

    def _resolve_json_metadata(
        self,
        template_metadata: dict,
        kwargs: dict,
        context: VulkanExecutionContext,
    ) -> dict:
        """Replace template expressions with actual node data using Jinja2."""
        try:
            env_config = getattr(context.resources, POLICY_CONFIG_KEY, None)
            env_variables = env_config.variables if env_config else {}
        except Exception:
            env_variables = {}

        return resolve_value(template_metadata, kwargs, env_variables)

    def _terminate(
        self,
        context: OpExecutionContext,
        result: str,
        metadata: dict[str, Any] | None = None,
    ) -> bool:
        app_client_resource: AppClientResource = getattr(
            context.resources, APP_CLIENT_KEY
        )
        client = app_client_resource.get_client()
        backend_run_id: str = context.run_id

        context.log.debug(f"Returning status {result} for Dagster run {backend_run_id}")

        success = client.update_run_status(
            status=RunStatus.COMPLETED.value,
            result=result,
            metadata=metadata,
        )

        if not success:
            context.log.error(
                f"Failed to return status {result} for Dagster run {backend_run_id}"
            )

        return success

    @classmethod
    def from_spec(cls, node: TerminateNode):
        dependencies = node.dependencies

        return cls(
            name=normalize_node_id(node.id),
            description=node.description,
            return_status=node.return_status,
            output_data=node.output_data,
            dependencies=dependencies,
            callback=node.callback,
        )


class DagsterBranch(BranchNode, DagsterNode):
    def __init__(
        self,
        name: str,
        description: str,
        func: Callable,
        choices: list[str],
        dependencies: dict[str, Any],
    ):
        super().__init__(
            name=name,
            description=description,
            func=func,
            choices=choices,
            dependencies=dependencies,
        )
        self._func = _with_vulkan_context(self.func)

    def op(self) -> OpDefinition:
        def fn(context, inputs):
            inputs = _resolved_inputs(inputs, self.dependencies)
            try:
                output = self._func(context, **inputs)
                yield Output(None, output)
            except Exception as e:
                raise UserCodeException(self.name) from e

        branch_paths = {out: Out(is_required=False) for out in self.choices}
        node_op = OpDefinition(
            compute_fn=fn,
            name=self.name,
            ins={k: In() for k in self.dependencies.keys()},
            outs={
                **branch_paths,
            },
            required_resource_keys={POLICY_CONFIG_KEY},
        )
        return node_op

    @classmethod
    def from_spec(cls, node: BranchNode):
        return cls(
            name=normalize_node_id(node.id),
            description=node.description,
            func=node.func,
            choices=node.choices,
            dependencies=node.dependencies,
        )


class DagsterInput(InputNode, DagsterNode):
    def __init__(self, description: str, schema: dict[str, type], name="input_node"):
        super().__init__(name=name, description=description, schema=schema)

    def run(self, context, *args, **kwargs):
        config = context.op_config
        yield Output(config)

    def op(self):
        return OpDefinition(
            compute_fn=self.run,
            name=self.name,
            ins={},
            outs={"result": Out()},
            config_schema=self.schema,
        )

    @classmethod
    def from_spec(cls, node: InputNode):
        return cls(
            name=normalize_node_id(node.id),
            description=node.description,
            schema=node.schema,
        )


class DagsterConnection(ConnectionNode, DagsterNode):
    def __init__(
        self,
        name: str,
        url: str,
        method: str = "GET",
        description: str | None = None,
        headers: dict | None = None,
        params: dict | None = None,
        body: dict | None = None,
        timeout: int | None = None,
        retry_max_retries: int = 1,
        response_type: str = "JSON",
        dependencies: dict | None = None,
    ):
        super().__init__(
            name=name,
            url=url,
            method=method,
            description=description,
            headers=headers,
            params=params,
            body=body,
            timeout=timeout,
            retry_max_retries=retry_max_retries,
            response_type=response_type,
            dependencies=dependencies,
        )

    def op(self) -> OpDefinition:
        return OpDefinition(
            compute_fn=self.run,
            name=self.name,
            ins={k: In() for k in self.dependencies.keys()},
            outs={
                "result": Out(),
            },
            required_resource_keys={RUN_CONFIG_KEY, POLICY_CONFIG_KEY},
        )

    def run(self, context, inputs):
        env: VulkanPolicyConfig = getattr(context.resources, POLICY_CONFIG_KEY)
        inputs = _resolved_inputs(inputs, self.dependencies)

        try:
            config = HTTPConfig(
                url=self.url,
                method=self.method,
                headers=self.headers,
                params=self.params,
                body=self.body,
                retry=RetryPolicy(max_retries=self.retry_max_retries),
                response_type=self.response_type,
                timeout=self.timeout,
            )

            client = HTTPClient(config)
            response = client.execute_raw(inputs, env.variables)

            response.raise_for_status()

            if response.status_code == 200:
                result = format_response_data(response.content, self.response_type)
                yield Output(result)
        except httpx.HTTPError as e:
            context.log.error(f"Failed HTTP request: {str(e)}")
            raise e
        except Exception as e:
            context.log.error(str(e))
            raise e

    @classmethod
    def from_spec(cls, node: ConnectionNode):
        return cls(
            name=normalize_node_id(node.id),
            url=node.url,
            method=node.method,
            description=node.description,
            headers=node.headers,
            params=node.params,
            body=node.body,
            timeout=node.timeout,
            retry_max_retries=node.retry_max_retries,
            response_type=node.response_type,
            dependencies=node.dependencies,
        )


class DagsterDecision(DecisionNode, DagsterNode):
    def __init__(
        self,
        name: str,
        description: str,
        conditions: list[DecisionCondition],
        dependencies: dict[str, Any],
    ):
        super().__init__(
            name=name,
            description=description,
            conditions=conditions,
            dependencies=dependencies,
        )

    def _decision_fn(self, context, inputs):
        if_cond = next(c for c in self.conditions if c.decision_type == DecisionType.IF)
        else_cond = next(
            c for c in self.conditions if c.decision_type == DecisionType.ELSE
        )
        elif_conds = [
            c for c in self.conditions if c.decision_type == DecisionType.ELSE_IF
        ]
        if evaluate_condition(if_cond.condition, inputs):
            return if_cond.output
        for elif_cond in elif_conds:
            if evaluate_condition(elif_cond.condition, inputs):
                return elif_cond.output
        return else_cond.output

    def op(self) -> OpDefinition:
        def fn(context, inputs):
            inputs = _resolved_inputs(inputs, self.dependencies)
            try:
                output = self._decision_fn(context, inputs)
                yield Output(None, output)
            except Exception as e:
                raise UserCodeException(self.name) from e

        branch_paths = {
            condition.output: Out(is_required=False) for condition in self.conditions
        }
        node_op = OpDefinition(
            compute_fn=fn,
            name=self.name,
            ins={k: In() for k in self.dependencies.keys()},
            outs={
                **branch_paths,
            },
            required_resource_keys={POLICY_CONFIG_KEY},
        )
        return node_op

    @classmethod
    def from_spec(cls, node: DecisionNode):
        return cls(
            name=normalize_node_id(node.id),
            description=node.description,
            conditions=node.conditions,
            dependencies=node.dependencies,
        )


_NODE_TYPE_MAP: dict[type[Node], type[DagsterNode]] = {
    TransformNode: DagsterTransform,
    TerminateNode: DagsterTerminate,
    BranchNode: DagsterBranch,
    DataInputNode: DagsterDataInput,
    InputNode: DagsterInput,
    PolicyDefinitionNode: DagsterPolicy,
    ConnectionNode: DagsterConnection,
    DecisionNode: DagsterDecision,
}


def to_dagster_nodes(nodes: list[Node]) -> list[DagsterNode]:
    return [to_dagster_node(node) for node in nodes]


def to_dagster_node(node: Node) -> DagsterNode:
    typ = type(node)
    impl_type = _NODE_TYPE_MAP.get(typ)
    if impl_type is None:
        msg = f"Node type {typ} has no known Dagster implementation"
        raise ValueError(msg)

    return impl_type.from_spec(node)
